from musc.evaluator.high_level import Evaluator
from musc.hpo.distr.high_level import SingleValue, into_distr
from musc.hpo.high_level.a import UpdateStrategySearch
from musc.service.concepts.a import BaseModel, Metric
from musc.service.concepts.high_level.a import (
    UpdateStrategyByDriftDetection,
    UpdateStrategyByPeriodicallyUpdate,
    UpdateStrategyNone,
)
from musc.service.high_level import Service, run_service_http

from . import drift_detectors

__all__ = [
    'BaseModel',
    'Evaluator',
    'Metric',
    'Service',
    'SingleValue',
    'UpdateStrategyByDriftDetection',
    'UpdateStrategyByPeriodicallyUpdate',
    'UpdateStrategyNone',
    'UpdateStrategySearch',
    'drift_detectors',
    'into_distr',
    'run_service_http',
]

try:
    import torch as _
except ImportError:
    pass
else:
    from musc.adaptors.models import ModelPyTorch
    from musc.hpo.res_pool.torch_device import TorchDevicePool
    __all__.extend(['ModelPyTorch', 'TorchDevicePool'])

try:
    import sklearn as _
except ImportError:
    pass
else:
    from musc.adaptors.models import ModelScikitLearn
    __all__.extend(['ModelScikitLearn'])



from __future__ import annotations

from collections.abc import Callable
from typing import Any, Generic, TypeVar

from river.base import BinaryDriftDetector, DriftDetector

from musc.adaptors.drift_detectors.a import DriftDetectorNone, DriftDetectorPeriodicallyUpdate
from musc.adaptors.drift_detectors.river.a import (
    DriftDetectorByRiver,
    DriftDetectorByRiverBinary,
    DriftDetectorByRiverBinaryWithNcl,
    DriftDetectorByRiverWithNcl,
)
from musc.adaptors.drift_detectors.river.with_kdp import ADWIN, PageHinkley, kswin_distr_prelude
from musc.hpo.distr.a import BaseDistribution
from musc.hpo.distr.high_level import into_distr
from musc.service.concepts.a import BaseModel, Metric, UpdateStrategy


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')


class ModelGuard(Generic[XR, X, YR, Y, YP], BaseModel[XR, X, YR, Y, YP]):

    def predict(self, x: X) -> YP:
        del x
        raise RuntimeError

    def inner(self) -> Any:
        raise RuntimeError

    def clone(self) -> BaseModel[XR, X, YR, Y, YP]:
        raise RuntimeError

    def clone_with_resource(self, resource: Any) -> BaseModel[XR, X, YR, Y, YP]:
        del resource
        raise RuntimeError

    def resource_current(self) -> Any:
        raise RuntimeError

    def resource_cpu(self) -> Any:
        raise RuntimeError

    def preprocess_x(self, x: XR) -> X:
        del x
        raise RuntimeError

    def preprocess_y(self, y: YR) -> Y:
        del y
        raise RuntimeError

    def y_pred_to_json_like(self, y_pred: YP) -> Any:
        del y_pred
        raise RuntimeError


class UpdateStrategySubclass(UpdateStrategy):

    def __repr__(self) -> str:
        return UpdateStrategy(
            self.drift_detector,
            self.updator,
            self.data_amount_required,
            self.use_ncl,
        ).__repr__()


class UpdateStrategyNone(UpdateStrategySubclass):

    def __init__(self) -> None:
        super().__init__(DriftDetectorNone(), updator_no_op, 0, False)

    def clone_without_state(self) -> UpdateStrategyNone:
        return UpdateStrategyNone()


class UpdateStrategyByPeriodicallyUpdate(UpdateStrategySubclass):

    def __init__(
        self,
        period: int,
        updator: Callable[[Any, list[Any], list[Any]], None],
        *,
        ncl: int | None = None,
        probability: float | None = None,
    ) -> None:
        super().__init__(
            DriftDetectorPeriodicallyUpdate(period, ncl, probability),
            UpdatorWrapper(updator),
            period,
            True,
        )


class UpdateStrategyByDriftDetection(UpdateStrategySubclass):

    def __init__(
        self,
        drift_detector: BinaryDriftDetector | DriftDetector,
        updator: Callable[[Any, list[Any], list[Any]], None],
        data_amount_required: int,
        metric: Metric | None = None,
        use_ncl: bool | None = None,
    ) -> None:
        if isinstance(drift_detector, BinaryDriftDetector):
            if not hasattr(drift_detector, 'new_concept_length'):
                drift_detector_ = DriftDetectorByRiverBinary(drift_detector)
            else:
                drift_detector_ = DriftDetectorByRiverBinaryWithNcl(drift_detector)
        elif isinstance(drift_detector, DriftDetector):
            if metric is None:
                raise TypeError
            if not hasattr(drift_detector, 'new_concept_length'):
                drift_detector_ = DriftDetectorByRiver(drift_detector, metric)
            else:
                drift_detector_ = DriftDetectorByRiverWithNcl(drift_detector, metric)
        else:
            raise TypeError
        drift_detector_ = drift_detector_.clone_without_state()
        updator_ = UpdatorWrapper(updator)
        use_ncl = use_ncl if use_ncl is not None else hasattr(drift_detector, 'new_concept_length')
        super().__init__(drift_detector_, updator_, data_amount_required, use_ncl)

    @staticmethod
    def kwargs_distr_prelude() -> dict[str, BaseDistribution[Any]]:
        return {
            'drift_detector': into_distr([
                {'type': ADWIN},
                kswin_distr_prelude(),
                {'type': PageHinkley},
            ]),
        }


class UpdatorWrapper:

    def __init__(self, inner: Callable[[Any, list[Any], list[Any]], None]) -> None:
        self._inner = inner

    def __call__(self, model: BaseModel, x_arr: list[Any], y_arr: list[Any]) -> None:
        self._inner(model.inner(), x_arr, y_arr)


def apply_model_adaptors(model: Any) -> BaseModel:

    if isinstance(model, BaseModel):
        return model

    try:
        import torch
    except ImportError:
        pass
    else:
        from musc.adaptors.models import ModelPyTorch
        if isinstance(model, torch.nn.Module):
            return ModelPyTorch(model)

    try:
        import sklearn
    except ImportError:
        pass
    else:
        from musc.adaptors.models import ModelScikitLearn
        if isinstance(model, sklearn.base.BaseEstimator):
            return ModelScikitLearn(model)

    raise TypeError('The model is not recognized by musc')


def updator_no_op(model: BaseModel, x_arr: list[Any], y_arr: list[Any]) -> None:
    del model, x_arr, y_arr


__all__ = [
    'ModelGuard',
    'UpdateStrategyNone',
    'UpdateStrategyByPeriodicallyUpdate',
    'UpdateStrategyByDriftDetection',
    'apply_model_adaptors',
]



from collections.abc import Callable
from io import StringIO
from typing import Any, Generic, Hashable, TypeVar

from musc.commons import explain_object
from musc.hpo.distr.a import (
    BaseDistribution,
    DistributionChoice,
    DistributionJoint,
    DistributionOnePoint,
    DistributionScipyWrapper,
)


V = TypeVar('V')


def into_distr(obj: Any) -> BaseDistribution[Any]:
    if isinstance(obj, BaseDistribution):
        return obj
    if hasattr(obj, 'rvs'):
        return DistributionScipyWrapper(obj)
    if isinstance(obj, list):
        return DistributionChoice([into_distr(obj_) for obj_ in obj])
    if isinstance(obj, dict):
        obj = dict(obj)
        if 'type' in obj:
            return DistributionJointType(
                obj.pop('type'),
                {k: into_distr(v) for k, v in obj.items()},
            )
        if 'base_fn' in obj:
            return DistributionJointBaseFn(
                obj.pop('base_fn'),
                {k: into_distr(v) for k, v in obj.items()},
            )
        raise ValueError
    return DistributionOnePoint(obj)


class DistributionJointType(Generic[V], DistributionJoint[V]):

    def __init__(
        self,
        type_: Callable[..., V],
        kwargs: dict[str, BaseDistribution[Any]],
    ) -> None:
        super().__init__(type_, kwargs=kwargs)


class DistributionJointBaseFn(DistributionJoint[Any]):

    def __init__(
        self,
        base_fn: Callable[..., Any],
        kwargs: dict[str, BaseDistribution[Any]],
    ):
        super().__init__(lambda **ka_: lambda *a, **ka: base_fn(*a, **ka, **ka_), kwargs=kwargs)
        self._base_fn = base_fn

    def explain_trace(self, trace: tuple[Hashable, ...]) -> str:
        expr = StringIO()
        expr.write(f'(lambda *args, **kwargs: {explain_object(self._base_fn)}(*args, **kwargs')
        for (k, d), st in zip(self._kwargs.items(), trace[len(self._args):]):
            expr.write(f', {k}={d.explain_trace(st)}')
        expr.write('))')
        return expr.getvalue()


class SingleValue(Generic[V], DistributionOnePoint[V]):

    pass


__all__ = [
    'into_distr',
    'SingleValue',
]



from typing import Generic, TypeVar

from musc.service.concepts.a import BaseModel
from musc.service.concepts.high_level.a import ModelGuard


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')


class ModelNoClone(Generic[XR, X, YR, Y, YP], ModelGuard[XR, X, YR, Y, YP]):

    def __init__(self, inner: BaseModel[XR, X, YR, Y, YP]) -> None:
        self.__inner = inner

    def clone(self) -> BaseModel[XR, X, YR, Y, YP]:
        return self.__inner


def no_clone(model: BaseModel[XR, X, YR, Y, YP]) -> ModelNoClone[XR, X, YR, Y, YP]:
    return ModelNoClone(model)


__all__ = [
    'no_clone',
]



from __future__ import annotations

import itertools
import logging
import re
import time
import typing
from collections.abc import Callable, Hashable, Iterable, Sequence
from multiprocessing import Manager
from os import PathLike
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Any, Generic, Literal, TypeVar

import pandas as pd

from musc.commons import safe_save
from musc.evaluator.high_level import Evaluator
from musc.hpo.a import ExhaustiveSearch, RandomizedSearch, Sample
from musc.hpo.distr.a import BaseDistribution
from musc.hpo.distr.high_level import into_distr
from musc.hpo.high_level.helper import ModelCloneHelper
from musc.hpo.res_pool.base import BaseResourcePool
from musc.hpo.res_pool.global_ import GlobalResourcePool
from musc.service.concepts.a import BaseModel, Metric, UpdateStrategy
from musc.service.concepts.high_level.a import apply_model_adaptors
from musc.service.impl.dd_process.stats import Stats


S = TypeVar('S')


class BaseUpdateStrategySearch(Generic[S]):

    def __init__(
        self,
        distr: BaseDistribution[UpdateStrategy],
        model: BaseModel,
        x_arr: Iterable[Any],
        y_arr: Iterable[Any],
        t_x_arr: Iterable[Any],
        t_y_arr: Iterable[Any],
        metric: list[Metric],
        score_fn: Callable[[list[float], Stats], S],
        greater_score_fn: Callable[[S, S], bool] | None,
        weak_greater_score_fn: Callable[[S, S], bool] | None,
        eval_retry: int,
        top_k_kept: int,
        exhaustive: bool,
        n_jobs: int | None,
        resource_pool: BaseResourcePool[Any] | None,
        sample_callback: Callable[[Sample[UpdateStrategy, None]], None] | None,
        optimal_candidates_callback: Callable[
            [
                Sample[UpdateStrategy, S],
                list[Sample[UpdateStrategy, S]],
                set[Hashable] | None,
                int | None,
            ],
            bool,
        ] | None,
        suboptimal_candidates_callback: Callable[
            [Sample[UpdateStrategy, S], bool, set[Hashable] | None, int | None],
            bool,
        ] | None,
        top_k_callback: \
            Callable[[Sample[UpdateStrategy, S], list[Sample[UpdateStrategy, S]]], None] | None,
        debug: bool,
    ) -> None:

        eval_max_n_tries = eval_retry + 1 if eval_retry >= 0 else None

        # Sending CUDA model to other processes is problematic.
        # See https://pytorch.org/docs/stable/multiprocessing.html#sharing-cuda-tensors
        # for details.
        resource_initial = model.resource_current()
        self._model_clone_helper = ModelCloneHelper(model.clone_with_resource(model.resource_cpu()))
        model = self._model_clone_helper.portable()

        if resource_pool is not None:
            self._pool = GlobalResourcePool(resource_pool)
            global_resource_pool = self._pool.portable()
        else:
            self._pool = global_resource_pool = None

        def score_fn_(us: UpdateStrategy) -> S:
            if not isinstance(us, UpdateStrategy):
                raise TypeError
            score, stats = None, None
            for i in range(eval_max_n_tries) if eval_max_n_tries is not None else itertools.count():
                try:
                    if global_resource_pool is None:
                        score, stats = Evaluator(model.clone_with_resource(resource_initial), us) \
                            .evaluate(x_arr, y_arr, t_x_arr, t_y_arr, metric)
                    else:
                        with global_resource_pool.alloc() as resource:
                            score, stats = Evaluator(model.clone_with_resource(resource), us) \
                                .evaluate(x_arr, y_arr, t_x_arr, t_y_arr, metric)
                except Exception as e:
                    if debug:
                        raise
                    eval_max_n_tries_ = eval_max_n_tries if eval_max_n_tries is not None else 'inf'
                    logging.warn(
                        f'The evaluation process raised an exception: {repr(e)}'
                        f' (number of tries: {i+1}/{eval_max_n_tries_})'
                    )
                    time.sleep(60.0)
                else:
                    break
            if score is None or stats is None:
                raise RuntimeError('Failed to execute evaluation')
            return score_fn(score, stats)

        if not exhaustive:
            if optimal_candidates_callback is not None:
                optimal_candidates_callback_ \
                    = lambda s, o, t: optimal_candidates_callback(s, o, t, None)
            else:
                optimal_candidates_callback_ = None
            if suboptimal_candidates_callback is not None:
                suboptimal_candidates_callback_ \
                    = lambda s, n, t: suboptimal_candidates_callback(s, n, t, None)
            else:
                suboptimal_candidates_callback_ = None
            self._inner = RandomizedSearch(
                distr,
                score_fn_,
                greater_score_fn,
                weak_greater_score_fn,
                top_k_kept,
                n_jobs,
                sample_callback,
                optimal_candidates_callback_,
                suboptimal_candidates_callback_,
                top_k_callback,
            )
        else:
            if optimal_candidates_callback is not None:
                optimal_candidates_callback_ \
                    = lambda s, o, c: optimal_candidates_callback(s, o, None, c)
            else:
                optimal_candidates_callback_ = None
            if suboptimal_candidates_callback is not None:
                suboptimal_candidates_callback_ \
                    = lambda s, n, c: suboptimal_candidates_callback(s, n, None, c)
            else:
                suboptimal_candidates_callback_ = None
            self._inner = ExhaustiveSearch(
                distr,
                score_fn_,
                greater_score_fn,
                weak_greater_score_fn,
                top_k_kept,
                n_jobs,
                sample_callback,
                optimal_candidates_callback_,
                suboptimal_candidates_callback_,
                top_k_callback,
            )

    def search(self, n_iter: int) -> bool:
        return self._inner.search(n_iter)

    def optimal_candidates(self) -> list[Sample[UpdateStrategy, S]]:
        return self._inner.survivors()

    def top_k(self) -> list[Sample[UpdateStrategy, S]]:
        return self._inner.top_k()

    def __enter__(self) -> BaseUpdateStrategySearch[S]:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        del exc_type, exc_value, traceback
        if self._pool is not None:
            self._pool.stop()
        self._model_clone_helper.stop()


class UpdateStrategySearch(BaseUpdateStrategySearch[list[Any]]):

    def __init__(
        self,
        search_space: Any,
        model: Any,
        x_arr: Iterable[Any],
        y_arr: Iterable[Any],
        t_x_arr: Iterable[Any],
        t_y_arr: Iterable[Any],
        metric: Metric | Sequence[Metric | Literal['time', 'n_updates', 'n_samples']],
        optim_mode: Literal['min', 'max'] | Sequence[Literal['min', 'max']] | None,
        *,
        weight_fn: Callable[..., Any] | None = None,
        top_k_kept: int = 0,
        exhaustive: bool = False,
        n_jobs: int | None = None,
        resource_pool: BaseResourcePool[Any] | None = None,
        verbose_level: Literal[0, 1, 2] = 2,
        trace_history_file_path: str | PathLike | None = None,
        exhaustive_count_file_path: str | PathLike | None = None,
        optimal_scores_csv_path: str | PathLike | None = None,
        optimal_samples_file_path: str | PathLike | None = None,
        top_k_scores_csv_path: str | PathLike | None = None,
        top_k_samples_file_path: str | PathLike | None = None,
        load_old_state: bool = False,
        stop_signal_file_path: str | PathLike | None = None,
        eval_retry: int = 0,
        debug: bool = False,
        sample_callback: Callable[[Sample[UpdateStrategy, None]], None] | None = None,
        optimal_candidates_callback: Callable[
            [
                Sample[UpdateStrategy, list[Any]],
                list[Sample[UpdateStrategy, list[Any]]],
                set[Hashable] | None,
                int | None,
            ],
            bool,
        ] | None = None,
        suboptimal_candidates_callback: Callable[
            [Sample[UpdateStrategy, list[Any]], bool, set[Hashable] | None, int | None],
            bool,
        ] | None = None,
        top_k_callback: Callable[
            [Sample[UpdateStrategy, list[Any]], list[Sample[UpdateStrategy, list[Any]]]],
            None,
        ] | None = None,
    ) -> None:

        metric_: list[Metric | Literal['time', 'n_updates', 'n_samples']]
        match metric:
            case Metric():
                metric_ = [metric]
                metric__ = [metric]
            case Sequence():
                metric_ = list(metric)
                metric__ = [m for m in metric_ if isinstance(m, Metric)]
        del metric

        optim_mode_: list[Literal['min', 'max']] | None
        match optim_mode:
            case None:
                optim_mode_ = None
            case 'min':
                optim_mode_ = ['min']
            case 'max':
                optim_mode_ = ['max']
            case _:
                optim_mode_ = list(optim_mode)
        del optim_mode

        def score_fn(score: list[float], stats: Stats) -> list[Any]:
            score_and_stats = list[float | int]()
            score_i = 0
            for m in metric_:
                match m:
                    case Metric():
                        score_and_stats.append(score[score_i])
                        score_i += 1
                    case 'time':
                        score_and_stats.append(stats.model_update_cost_by_time())
                    case 'n_updates':
                        score_and_stats.append(stats.model_update_cost_by_num_updates())
                    case 'n_samples':
                        score_and_stats.append(stats.model_update_cost_by_num_samples())
            if weight_fn is None:
                return score_and_stats
            weighted = weight_fn(*score_and_stats)
            if isinstance(weighted, Sequence):
                return list(weighted)
            return [weighted]

        if optim_mode_ is not None:
            def greater_score_fn(s1: list[Any], s2: list[Any]) -> bool:
                if not (len(s1) == len(s2) == len(optim_mode_)):
                    raise RuntimeError
                greater = False
                for s1_, s2_, o in zip(s1, s2, optim_mode_, strict=True):
                    match o:
                        case 'min':
                            if s1_ > s2_:
                                return False
                            if s1_ < s2_:
                                greater = True
                        case 'max':
                            if s1_ < s2_:
                                return False
                            if s1_ > s2_:
                                greater = True
                return greater
            greater_score_fn_ = greater_score_fn
        else:
            greater_score_fn_ = None

        if optim_mode_ is not None:
            def weak_greater_score_fn(s1: list[Any], s2: list[Any]) -> bool:
                if not (len(s1) == len(s2) == len(optim_mode_)):
                    raise RuntimeError
                for s1_, s2_, o in zip(s1, s2, optim_mode_, strict=True):
                    match o:
                        case 'min':
                            if s1_ < s2_:
                                return True
                            if s1_ > s2_:
                                return False
                        case 'max':
                            if s1_ > s2_:
                                return True
                            if s1_ < s2_:
                                return False
                return False
            weak_greater_score_fn_ = weak_greater_score_fn
        else:
            weak_greater_score_fn_ = None

        multiprocessing_manager = Manager()
        log_queue = multiprocessing_manager.Queue()
        log_thread = Thread(target=self._log_thread_target, args=(log_queue,), daemon=True)
        log_thread.start()

        if sample_callback is None:
            def sample_callback_(sample: Sample[UpdateStrategy, None]) -> None:
                if verbose_level >= 1:
                    log_queue.put({'event': 'evaluating', 'sample': sample.explain()})
            sample_callback__ = sample_callback_
        else:
            sample_callback__ = sample_callback

        if optimal_candidates_callback is None:
            def optimal_candidates_callback_(
                sample: Sample[UpdateStrategy, list[Any]],
                optimal_candidates_now: list[Sample[UpdateStrategy, list[Any]]],
                trace_history_now: set[Hashable] | None,
                exhaustive_count_now: int | None,
            ) -> bool:
                self._log_event(
                    sample,
                    optimal_candidates_now,
                    verbose_level,
                    'optimal',
                    'optimal_candidates_now',
                    log_queue,
                )
                if trace_history_file_path is not None:
                    if trace_history_now is not None:
                        self._write_traces(trace_history_now, trace_history_file_path)
                if exhaustive_count_file_path is not None:
                    if exhaustive_count_now is not None:
                        self._write_exhaustive_count(
                            exhaustive_count_now,
                            exhaustive_count_file_path,
                        )
                if optimal_scores_csv_path is not None:
                    self._write_scores(
                        optimal_candidates_now,
                        optimal_scores_csv_path,
                        metric_,
                        weight_fn,
                    )
                if optimal_samples_file_path is not None:
                    self._write_samples(optimal_candidates_now, optimal_samples_file_path)
                if stop_signal_file_path is not None and Path(stop_signal_file_path).is_file():
                    Path(stop_signal_file_path).unlink()
                    return True
                return False
            optimal_candidates_callback__ = optimal_candidates_callback_
        else:
            optimal_candidates_callback__ = optimal_candidates_callback

        if suboptimal_candidates_callback is None:
            def suboptimal_candidates_callback_(
                sample: Sample[UpdateStrategy, list[Any]],
                is_new: bool,
                trace_history_now: set[Hashable] | None,
                exhaustive_count_now: int | None,
            ) -> bool:
                if verbose_level >= 1:
                    log_queue.put({
                        'event': 'suboptimal',
                        'sample': sample.explain(),
                        'is_new': is_new,
                        'score': sample.score if len(sample.score) != 1 else sample.score[0],
                    })
                if is_new:
                    if trace_history_file_path is not None:
                        if trace_history_now is not None:
                            self._write_traces(trace_history_now, trace_history_file_path)
                    if exhaustive_count_file_path is not None:
                        if exhaustive_count_now is not None:
                            self._write_exhaustive_count(
                                exhaustive_count_now,
                                exhaustive_count_file_path,
                            )
                    if stop_signal_file_path is not None and Path(stop_signal_file_path).is_file():
                        Path(stop_signal_file_path).unlink()
                        return True
                return False
            suboptimal_candidates_callback__ = suboptimal_candidates_callback_
        else:
            suboptimal_candidates_callback__ = suboptimal_candidates_callback

        if top_k_callback is None:
            def top_k_callback_(
                sample: Sample[UpdateStrategy, list[Any]],
                top_k_now: list[Sample[UpdateStrategy, list[Any]]],
            ) -> None:
                self._log_event(
                    sample,
                    top_k_now,
                    verbose_level,
                    'top_k',
                    'top_k_now',
                    log_queue,
                )
                if top_k_scores_csv_path is not None:
                    self._write_scores(top_k_now, top_k_scores_csv_path, metric_, weight_fn)
                if top_k_samples_file_path is not None:
                    self._write_samples(top_k_now, top_k_samples_file_path)
            top_k_callback__ = top_k_callback_
        else:
            top_k_callback__ = top_k_callback

        super().__init__(
            into_distr(search_space),
            apply_model_adaptors(model),
            x_arr,
            y_arr,
            t_x_arr,
            t_y_arr,
            metric__,
            score_fn,
            greater_score_fn_,
            weak_greater_score_fn_,
            eval_retry,
            top_k_kept,
            exhaustive,
            n_jobs,
            resource_pool,
            sample_callback__,
            optimal_candidates_callback__,
            suboptimal_candidates_callback__,
            top_k_callback__,
            debug,
        )
        self._multiprocessing_manager = multiprocessing_manager
        self._log_queue = log_queue
        self._log_thread = log_thread

        if load_old_state:
            self._load_old_state(
                exhaustive,
                trace_history_file_path,
                exhaustive_count_file_path,
                optimal_scores_csv_path,
                optimal_samples_file_path,
                top_k_scores_csv_path,
                top_k_samples_file_path,
            )

    @staticmethod
    def _log_thread_target(log_queue: Queue[Any]) -> None:
        while (info := log_queue.get()) is not None:
            logging.info(info)

    @staticmethod
    def _log_event(
        cur_sample: Sample[UpdateStrategy, list[Any]],
        all_samples: list[Sample[UpdateStrategy, list[Any]]],
        verbose_level: Literal[0, 1, 2],
        event_name: str,
        all_samples_name: str,
        log_queue: Queue[Any],
    ) -> None:
        if verbose_level >= 1:
            info = {
                'event': event_name,
                'sample': cur_sample.explain(),
                'score': cur_sample.score if len(cur_sample.score) != 1 else cur_sample.score[0],
            }
            if verbose_level >= 2:
                info[all_samples_name] = [
                    {
                        'sample': s.explain(),
                        'score': s.score if len(s.score) != 1 else s.score[0],
                    }
                    for s in all_samples
                ]
            log_queue.put(info)

    @staticmethod
    def _write_traces_unsafe(traces: set[Hashable], file_path: str | PathLike) -> None:
        with open(file_path, 'w') as f:
            for t in traces:
                print(repr(t), file=f)

    @staticmethod
    def _write_traces(traces: set[Hashable], file_path: str | PathLike) -> None:
        return safe_save(lambda p: UpdateStrategySearch._write_traces_unsafe(traces, p), file_path)

    @staticmethod
    def _write_exhaustive_count_unsafe(exhaustive_count: int, file_path: str | PathLike) -> None:
        Path(file_path).write_text(str(exhaustive_count))

    @staticmethod
    def _write_exhaustive_count(exhaustive_count: int, file_path: str | PathLike) -> None:
        return safe_save(
            lambda p: UpdateStrategySearch._write_exhaustive_count_unsafe(exhaustive_count, p),
            file_path,
        )

    @staticmethod
    def _write_scores(
        samples: list[Sample[UpdateStrategy, list[Any]]],
        csv_path: str | PathLike,
        metrics: list[Metric | Literal['time', 'n_updates', 'n_samples']],
        weight_fn: Callable[..., Any] | None,
    ) -> None:
        if len(samples) == 0:
            raise ValueError
        scores = [s.score for s in samples]
        if weight_fn is None:
            scores_columns = list[str]()
            metric_i = 0
            for m in metrics:
                match m:
                    case Metric():
                        scores_columns.append(f'metric_{metric_i}')
                        metric_i += 1
                    case _:
                        scores_columns.append(m)
        else:
            scores_columns = [f'weighted_{i}' for i in range(len(scores[0]))]
        safe_save(
            pd.DataFrame(scores, columns=scores_columns).rename_axis(index='i').to_csv,
            csv_path,
        )

    @staticmethod
    def _write_samples_unsafe(
        samples: list[Sample[UpdateStrategy, list[Any]]],
        file_path: str | PathLike,
    ) -> None:
        with open(file_path, 'w') as f:
            for i, s in enumerate(samples):
                i = str(i).rjust(len(str(len(samples) - 1)))
                print(f'{i}: {s.explain()} ##BEGIN#TRACE##{repr(s.trace)}##END#TRACE##', file=f)

    @staticmethod
    def _write_samples(
        samples: list[Sample[UpdateStrategy, list[Any]]],
        file_path: str | PathLike,
    ) -> None:
        return safe_save(
            lambda p: UpdateStrategySearch._write_samples_unsafe(samples, p),
            file_path,
        )

    def _load_old_state(
        self,
        exhaustive: bool,
        trace_history_file_path: str | PathLike | None,
        exhaustive_count_file_path: str | PathLike | None,
        optimal_scores_csv_path: str | PathLike | None,
        optimal_samples_file_path: str | PathLike | None,
        top_k_scores_csv_path: str | PathLike | None,
        top_k_samples_file_path: str | PathLike | None,
    ) -> None:

        if optimal_scores_csv_path is None:
            raise ValueError
        if optimal_samples_file_path is None:
            raise ValueError
        if top_k_scores_csv_path is None:
            raise ValueError
        if top_k_samples_file_path is None:
            raise ValueError
        if not exhaustive and trace_history_file_path is None:
            raise ValueError
        if exhaustive and exhaustive_count_file_path is None:
            raise ValueError

        if all(map(self._check_non_existence, [
            trace_history_file_path if not exhaustive else exhaustive_count_file_path,
            optimal_scores_csv_path,
            optimal_samples_file_path,
            top_k_scores_csv_path,
            top_k_samples_file_path,
        ])):
            return

        optimal_scores_csv_path = self._check_existence(optimal_scores_csv_path)
        optimal_samples_file_path = self._check_existence(optimal_samples_file_path)
        top_k_scores_csv_path = self._check_existence(top_k_scores_csv_path)
        top_k_samples_file_path = self._check_existence(top_k_samples_file_path)

        optimal_candidates = zip(
            self._load_traces_from_samples_file(optimal_samples_file_path),
            self._load_scores_from_csv(optimal_scores_csv_path),
            strict=True,
        )
        top_k = zip(
            self._load_traces_from_samples_file(top_k_samples_file_path),
            self._load_scores_from_csv(top_k_scores_csv_path),
            strict=True,
        )

        if not exhaustive:
            trace_history_file_path = self._check_existence(trace_history_file_path)
            trace_history = {
                eval(l)
                for l in Path(trace_history_file_path).read_text().splitlines()
            }
            typing.cast(RandomizedSearch[UpdateStrategy, list[Any]], self._inner).load_state(
                optimal_candidates,
                top_k,
                trace_history,
            )
        else:
            exhaustive_count_file_path = self._check_existence(exhaustive_count_file_path)
            exhaustive_count = eval(Path(exhaustive_count_file_path).read_text())
            typing.cast(ExhaustiveSearch[UpdateStrategy, list[Any]], self._inner).load_state(
                optimal_candidates,
                top_k,
                exhaustive_count,
            )

    @staticmethod
    def _check_non_existence(path: str | PathLike | None) -> bool:
        return path is not None and not Path(path).exists()

    @staticmethod
    def _check_existence(path: str | PathLike | None) -> str | PathLike:
        if path is None:
            raise ValueError
        if not Path(path).is_file():
            raise FileNotFoundError
        return path

    @staticmethod
    def _load_traces_from_samples_file(path: str | PathLike) -> list[Hashable]:
        traces = []
        for line in Path(path).read_text().splitlines():
            match = re.search('##BEGIN#TRACE##(.*)##END#TRACE##', line)
            if match is None:
                raise RuntimeError
            traces.append(eval(match.group(1)))
        return traces

    @staticmethod
    def _load_scores_from_csv(path: str | PathLike) -> list[list[Any]]:
        df = pd.read_csv(path)
        cols = [df[c].astype(object).values for c in df.columns[1:]]
        return [list(row) for row in zip(*cols, strict=True)]

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self._log_queue.put(None)
        self._log_thread.join()
        self._multiprocessing_manager.shutdown()
        return super().__exit__(exc_type, exc_value, traceback)


__all__ = [
    'UpdateStrategySearch',
]



from __future__ import annotations

import itertools
import random
import typing
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable
from io import StringIO
from typing import Any, Generic, Protocol, TypeVar

from musc.commons import explain_object


V, VH = TypeVar('V', covariant=True), TypeVar('VH', bound=Hashable)


class BaseDistribution(Generic[V], ABC):

    @abstractmethod
    def size(self) -> int | None:
        pass

    @abstractmethod
    def random_trace(self) -> Hashable:
        pass

    @abstractmethod
    def trace_by_index(self, index: int) -> Hashable:
        pass

    @abstractmethod
    def sample_by_trace(self, trace: Hashable) -> V:
        pass

    @abstractmethod
    def explain_trace(self, trace: Hashable) -> str:
        pass


class DistributionOnePoint(Generic[V], BaseDistribution[V]):

    def __init__(self, v: V) -> None:
        self._v = v

    def size(self) -> int | None:
        return 1

    def random_trace(self) -> None:
        return None

    def trace_by_index(self, index: int) -> None:
        if index != 0:
            raise IndexError
        return None

    def sample_by_trace(self, trace: None) -> V:
        del trace
        return self._v

    def explain_trace(self, trace: None) -> str:
        del trace
        return explain_object(self._v)


class DistributionChoice(Generic[V], BaseDistribution[V]):

    def __init__(
        self,
        population: list[BaseDistribution[V]],
        weights: list[float] | None = None,
    ) -> None:

        if len(population) == 0:
            raise ValueError

        self._population = population
        self._weights = weights

        self._size = 0
        self._cum_size = [0] * len(self._population)
        for i, d in enumerate(self._population):
            if (d_s := d.size()) is None:
                self._size = None
                self._cum_size = None
                break
            self._size += d_s
            for j in range(i+1, len(self._population)):
                self._cum_size[j] += d_s

    def size(self) -> int | None:
        return self._size

    def random_trace(self) -> tuple[int, Hashable]:
        index_in_population = random.choices(range(len(self._population)), self._weights)[0]
        return (index_in_population, self._population[index_in_population].random_trace())

    def trace_by_index(self, index: int) -> tuple[int, Hashable]:
        if self._size is None or self._cum_size is None or index >= self._size or index < 0:
            raise IndexError
        for i in range(len(self._population)-1, -1, -1):
            if index >= self._cum_size[i]:
                return (i, self._population[i].trace_by_index(index - self._cum_size[i]))
        assert False

    def sample_by_trace(self, trace: tuple[int, Hashable]) -> V:
        if trace[0] >= len(self._population) or trace[0] < 0:
            raise ValueError
        return self._population[trace[0]].sample_by_trace(trace[1])

    def explain_trace(self, trace: tuple[int, Hashable]) -> str:
        if trace[0] >= len(self._population) or trace[0] < 0:
            raise ValueError
        return self._population[trace[0]].explain_trace(trace[1])


class DistributionJoint(Generic[V], BaseDistribution[V]):

    def __init__(
        self,
        callable: Callable[..., V],
        args: list[BaseDistribution[Any]] | None = None,
        kwargs: dict[str, BaseDistribution[Any]] | None = None,
        include_prelude: bool = True,
    ) -> None:

        self._callable = callable
        self._args = args if args is not None else []
        if include_prelude and hasattr(callable, 'kwargs_distr_prelude'):
            self._kwargs = typing.cast(
                dict[str, BaseDistribution[Any]],
                callable.kwargs_distr_prelude(),
            )
        else:
            self._kwargs = {}
        if kwargs is not None:
            for k, d in kwargs.items():
                self._kwargs[k] = d

        self._size = 1
        for d in itertools.chain(self._args, self._kwargs.values()):
            if (d_s := d.size()) is None:
                self._size = None
                break
            self._size *= d_s

    def size(self) -> int | None:
        return self._size

    def random_trace(self) -> tuple[Hashable, ...]:
        return tuple(d.random_trace() for d in itertools.chain(self._args, self._kwargs.values()))

    def trace_by_index(self, index: int) -> tuple[Hashable, ...]:
        if self._size is None or index >= self._size or index < 0:
            raise IndexError
        trace = []
        for d in itertools.chain(self._args, self._kwargs.values()):
            assert isinstance(d_s := d.size(), int)
            index, i = divmod(index, d_s)
            trace.append(d.trace_by_index(i))
        return tuple(trace)

    def sample_by_trace(self, trace: tuple[Hashable, ...]) -> V:
        if len(trace) != len(self._args) + len(self._kwargs):
            raise ValueError
        args = [d.sample_by_trace(st) for d, st in zip(self._args, trace[:len(self._args)])]
        kwargs = {
            k: d.sample_by_trace(st)
            for (k, d), st in zip(self._kwargs.items(), trace[len(self._args):])
        }
        return self._callable(*args, **kwargs)

    def explain_trace(self, trace: tuple[Hashable, ...]) -> str:
        if len(trace) != len(self._args) + len(self._kwargs):
            raise ValueError
        expr = StringIO()
        expr.write(explain_object(self._callable))
        expr.write('(')
        write_comma = False
        for d, st in zip(self._args, trace[:len(self._args)]):
            if not write_comma:
                write_comma = True
            else:
                expr.write(', ')
            expr.write(d.explain_trace(st))
        for (k, d), st in zip(self._kwargs.items(), trace[len(self._args):]):
            if not write_comma:
                write_comma = True
            else:
                expr.write(', ')
            expr.write(f'{k}={d.explain_trace(st)}')
        expr.write(')')
        return expr.getvalue()


class DistributionScipyWrapper(Generic[VH], BaseDistribution[VH]):

    def __init__(self, inner: ProtocolScipyDistr[VH]) -> None:
        self._inner = inner

    def size(self) -> int | None:
        return None

    def random_trace(self) -> VH:
        return self._inner.rvs()

    def trace_by_index(self, index: int) -> VH:
        del index
        raise IndexError

    def sample_by_trace(self, trace: VH) -> VH:
        return trace

    def explain_trace(self, trace: VH) -> str:
        return explain_object(trace)


class ProtocolScipyDistr(Generic[V], Protocol):

    def rvs(self) -> V:
        raise NotImplementedError


__all__ = [
    'BaseDistribution',
    'DistributionOnePoint',
    'DistributionChoice',
    'DistributionJoint',
    'DistributionScipyWrapper',
]



from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import AbstractContextManager
from dataclasses import dataclass
from typing import Generic, TypeVar

from musc.service.impl.model_wrapper import ModelPredictionProcess, Prediction


X, Y, YP = TypeVar('X'), TypeVar('Y'), TypeVar('YP')


class BaseDataHistory(Generic[X, Y, YP], ABC):

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __getitem__(self, idx: int) -> DataHistoryItem[X, Y, YP]:
        pass

    @abstractmethod
    def gen_rlock(self) -> AbstractContextManager:
        pass

    @abstractmethod
    def recv_x(self, x: X, id_: str) -> YP:
        pass

    @abstractmethod
    def recv_y(self, y: Y, id_: str) -> None:
        pass

    @abstractmethod
    def register_listener(self, listener: BaseDataHistoryListener[X, Y, YP] | None) -> None:
        pass


@dataclass
class DataHistoryItem(Generic[X, Y, YP]):
    x: X
    y: Y | DataHistoryItemYNone
    y_pred: ModelPredictionProcess[X, YP] | Prediction[YP]


class DataHistoryItemYNone:
    pass


class BaseDataHistoryListener(Generic[X, Y, YP], ABC):

    @abstractmethod
    def notify_x(self, idx: int, x: X, y_pred: ModelPredictionProcess[X, YP]) -> None:
        pass

    @abstractmethod
    def notify_y(self, idx: int, y: Y) -> None:
        pass


__all__ = [
    'BaseDataHistory',
    'DataHistoryItem',
    'DataHistoryItemYNone',
    'BaseDataHistoryListener',
]



from __future__ import annotations

from collections.abc import Callable
from typing import Generic, TypeVar

from musc.evaluator.active_count import BaseActiveCount, NullActiveCount
from musc.service.concepts.a import BaseModel, UpdateStrategy
from musc.service.impl.model_wrapper import ModelWrapper
from musc.service.impl.data_history.base import BaseDataHistory
from musc.service.impl.data_history.in_memory import DataHistoryInMemory
from musc.service.impl.dd_process.a import DriftDetectorProcess
from musc.service.impl.dd_process.stats import Stats


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')
T = TypeVar('T')


class Service(Generic[XR, X, YR, Y, YP]):

    def __init__(
        self,
        model: BaseModel[XR, X, YR, Y, YP],
        update_strategy: UpdateStrategy[X, Y, YP],
        *,
        wait_y_timeout: float | None = None,
        daemon: bool = True,
        debug: bool = False,
        active_count: BaseActiveCount = NullActiveCount(),
    ) -> None:
        if not isinstance(active_count, NullActiveCount) and wait_y_timeout is not None:
            raise NotImplementedError(
                'Evaluating service with `wait_y_timeout` is currently not supported.'
            )
        self._model = ModelWrapper(model, update_strategy.updator, daemon, debug)
        self._data_history: BaseDataHistory[X, Y, YP] \
            = DataHistoryInMemory(self._model, active_count)
        self._drift_detector_process = DriftDetectorProcess(
            self._data_history,
            update_strategy,
            self._model,
            wait_y_timeout,
            daemon,
            active_count,
        )

    def recv_x(self, x: XR, id_: str) -> YP:
        return self._data_history.recv_x(self._model.preprocess_x(x), id_)

    def recv_y(self, y: YR, id_: str) -> None:
        return self._data_history.recv_y(self._model.preprocess_y(y), id_)

    def stats(self, fn: Callable[[Stats], T] = lambda s: s) -> T:
        return self._drift_detector_process.stats(fn)

    def stop_listening(self) -> None:
        self._drift_detector_process.stop_listening()

    def join(self) -> None:
        self._drift_detector_process.join()


__all__ = [
    'Service',
]



from __future__ import annotations

from copy import deepcopy
from typing import Any

import numpy as np

from musc.service.concepts.a import BaseModel


__all__ = []


try:

    import torch

except ImportError:

    pass

else:

    class ModelPyTorch(BaseModel):

        def __init__(self, model: torch.nn.Module) -> None:
            if (dtype := ModelPyTorch._determine_dtype(model)) is None:
                raise ValueError
            if (device := ModelPyTorch._determine_device(model)) is None:
                raise ValueError
            self._model = model
            self._dtype = dtype
            self._device = device

        def __del__(self) -> None:
            try:
                torch.cuda.empty_cache()  # To avoid memory leak
            except AttributeError:
                # AttributeError: 'NoneType' object has no attribute 'empty_cache'
                pass

        @staticmethod
        def _determine_dtype(model: torch.nn.Module) -> torch.dtype | None:
            dtype = None
            for p in model.parameters():
                if dtype is None:
                    dtype = p.dtype
                elif p.dtype != dtype:
                    return None
            return dtype

        @staticmethod
        def _determine_device(model: torch.nn.Module) -> torch.device | None:
            device = None
            for p in model.parameters():
                if device is None:
                    device = p.device
                elif p.device != device:
                    return None
            return device

        def predict(self, x: torch.Tensor) -> torch.Tensor:
            with torch.inference_mode():
                x = x.to(device=self._device)
                training_before = self._model.training
                self._model.eval()
                try:
                    return self._model(torch.stack([x]))[0]
                finally:
                    self._model.train(training_before)

        def inner(self) -> torch.nn.Module:
            return self._model

        def clone_with_resource(self, resource: Any) -> ModelPyTorch:
            if not isinstance(resource, torch.device):
                raise TypeError
            return ModelPyTorch(deepcopy(self._model).to(device=resource))

        def resource_current(self) -> Any:
            return self._device

        def resource_cpu(self) -> Any:
            return torch.device('cpu')

        def preprocess_x(self, x: Any) -> torch.Tensor:
            if isinstance(x, torch.Tensor):
                return x.to(dtype=self._dtype)
            return torch.tensor(x, dtype=self._dtype)

        def preprocess_y(self, y: Any) -> torch.Tensor:
            if isinstance(y, torch.Tensor):
                return y.to(dtype=self._dtype)
            return torch.tensor(y, dtype=self._dtype)

        def y_pred_to_json_like(self, y_pred: torch.Tensor) -> Any:
            return y_pred.tolist()

    __all__.append('ModelPyTorch')


try:

    import sklearn as _

except ImportError:

    pass

else:

    class ModelScikitLearn(BaseModel):

        def __init__(self, model: Any) -> None:
            self._model = model

        def predict(self, x: np.ndarray) -> np.ndarray:
            return self._model.predict(np.array([x]))[0]

        def inner(self) -> Any:
            return self._model

        def preprocess_x(self, x: Any) -> np.ndarray:
            # See https://numpy.org/doc/stable/reference/arrays.scalars.html#built-in-scalar-types
            # "The default data type in NumPy is float_."
            return np.array(x, dtype=np.float_)

        def preprocess_y(self, y: Any) -> np.ndarray:
            # See https://numpy.org/doc/stable/reference/arrays.scalars.html#built-in-scalar-types
            # "The default data type in NumPy is float_."
            return np.array(y, dtype=np.float_)

        def y_pred_to_json_like(self, y_pred: np.ndarray) -> Any:
            return y_pred.tolist()

    __all__.append('ModelScikitLearn')



from __future__ import annotations

from dataclasses import dataclass

from readerwriterlock.rwlock import RWLockFair


class StatsWithLock:

    def __init__(self) -> None:
        self._inner = Stats([])
        self._lock = RWLockFair()

    def inner(self) -> Stats:
        return self._inner

    def gen_rlock(self) -> RWLockFair._aReader:
        return self._lock.gen_rlock()

    def gen_wlock(self) -> RWLockFair._aWriter:
        return self._lock.gen_wlock()


@dataclass
class Stats:
    model_update_records: list[ModelUpdateRecord]
    def model_update_cost_by_time(self) -> float:
        return sum(r.time_spent for r in self.model_update_records)
    def model_update_cost_by_num_updates(self) -> int:
        return len(self.model_update_records)
    def model_update_cost_by_num_samples(self) -> int:
        return sum(
            r.new_concept_range[1] - r.new_concept_range[0]
            for r in self.model_update_records
        )


@dataclass
class ModelUpdateRecord:
    drift_point: int
    new_concept_range: tuple[int, int]
    time_spent: float


__all__ = [
    'Stats',
    'StatsWithLock',
    'ModelUpdateRecord',
]



from __future__ import annotations

import logging
from typing import Any, TypeVar

import flask
from flask import Flask
from gevent.pywsgi import WSGIServer

from musc.service.a import Service as ServiceLowLevel
from musc.service.errors import ServiceErrorByDuplicatedId, ServiceErrorFromModelPrediction
from musc.service.concepts.a import UpdateStrategy
from musc.service.concepts.high_level.a import apply_model_adaptors
from musc.service.impl.dd_process.stats import Stats


T = TypeVar('T')


class Service:

    def __init__(
        self,
        model: Any,
        update_strategy: UpdateStrategy,
        *,
        wait_y_timeout: float | None = None,
        daemon: bool = True,
        debug: bool = False,
    ) -> None:
        self._inner = ServiceLowLevel(
            apply_model_adaptors(model),
            update_strategy.clone_without_state(),
            wait_y_timeout=wait_y_timeout,
            daemon=daemon,
            debug=debug,
        )

    def recv_x(self, x: Any, id_: Any) -> Any:
        return self._inner.recv_x(x, str(id_))

    def recv_y(self, y: Any, id_: Any) -> None:
        return self._inner.recv_y(y, str(id_))

    def stats(self) -> Stats:
        return self._inner.stats()

    def stop_listening(self) -> None:
        return self._inner.stop_listening()

    def join(self) -> None:
        return self._inner.join()


def run_service_http(
    model: Any,
    update_strategy: UpdateStrategy,
    *,
    host: str = '127.0.0.1',
    port: int = 80,
    wait_y_timeout: float | None = None,
    debug: bool = False,
) -> None:

    model_ = apply_model_adaptors(model)
    service = Service(model_, update_strategy, wait_y_timeout=wait_y_timeout, debug=debug)

    flask_app = Flask(__name__)

    @flask_app.route('/', methods=['POST'])
    def recv_data() -> Any:

        request_json = flask.request.get_json(silent=True)
        if request_json is None:
            flask.abort(415)

        try:
            id_ = request_json['id']
        except KeyError:
            flask.abort(422)

        try:
            x = request_json['x']
        except KeyError:
            try:
                y = request_json['y']
            except KeyError:
                flask.abort(422)
            else:
                x, y, kind = None, model_.preprocess_y(y), 'y'
        else:
            try:
                y = request_json['y']
            except KeyError:
                x, y, kind = model_.preprocess_x(x), None, 'x'
            else:
                flask.abort(422)

        if kind == 'x':
            try:
                y_pred = service.recv_x(x, id_)
            except ServiceErrorByDuplicatedId:
                return {
                    'error':
                        f'Duplicated sample ID: {id_}. '
                        f'The service received an x-value with the same ID before.'
                }
            except ServiceErrorFromModelPrediction as e:
                basic_description = 'The underlying model prediction function raised an exception'
                logging.error(f'{basic_description}: {repr(e)}.')
                return {'error': f'{basic_description}.'}
            else:
                return {'y_pred': model_.y_pred_to_json_like(y_pred)}
        elif kind == 'y':
            try:
                service.recv_y(y, id_)
            except ServiceErrorByDuplicatedId:
                return {
                    'error':
                        f'Duplicated sample ID: {id_}. '
                        f'The service received a y-value with the same ID before.'
                }
            else:
                return {}
    del recv_data

    WSGIServer((host, port), flask_app).serve_forever()


__all__ = [
    'Service',
    'run_service_http',
]



from __future__ import annotations

import typing
from typing import Generic, TypeVar

from readerwriterlock.rwlock import RWLockFair

from musc.evaluator.active_count import BaseActiveCount
from musc.service.errors import ServiceErrorByDuplicatedId, ServiceErrorFromModelPrediction
from musc.service.impl.data_history.base import (
    BaseDataHistory,
    BaseDataHistoryListener,
    DataHistoryItem,
    DataHistoryItemYNone,
)
from musc.service.impl.model_wrapper import ModelPredictionProcess, ModelWrapper, Prediction


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')


class DataHistoryInMemory(Generic[X, Y, YP], BaseDataHistory[X, Y, YP]):

    def __init__(
        self,
        model: ModelWrapper[XR, X, YR, Y, YP],
        active_count: BaseActiveCount,
    ) -> None:

        self._model = model

        self._core = DataHistoryInMemoryCore[X, Y, YP]()
        self._id2idx = dict[str, int]()
        self._y_waiting_room = dict[str, Y]()

        self._listener: BaseDataHistoryListener[X, Y, YP] | None = None
        self._active_count = active_count

        self._lock = RWLockFair()

    def __len__(self) -> int:
        return len(self._core)

    def __getitem__(self, idx: int) -> DataHistoryItem[X, Y, YP]:
        return self._core[idx]

    def gen_rlock(self) -> RWLockFair._aReader:
        return self._lock.gen_rlock()

    def _gen_wlock(self) -> RWLockFair._aWriter:
        return self._lock.gen_wlock()

    def recv_x(self, x: X, id_: str) -> YP:
        with self._gen_wlock():
            if id_ in self._id2idx:
                raise ServiceErrorByDuplicatedId
            idx = len(self._core)
            y = self._y_waiting_room.pop(id_, DataHistoryItemYNone())
            y_pred = self._model.predict(x)
            self._id2idx[id_] = idx
            self._core.append(x, y, y_pred)
            listener = self._listener
        if listener is not None:
            listener.notify_x(idx, x, y_pred)
            if not isinstance(y, DataHistoryItemYNone):
                listener.notify_y(idx, y)
        y_pred = y_pred.join()
        with self._gen_wlock():
            self._core.recv_y_pred(idx, y_pred)
        if not y_pred.success:
            raise ServiceErrorFromModelPrediction(typing.cast(Exception, y_pred.y_pred))
        return typing.cast(YP, y_pred.y_pred)

    def recv_y(self, y: Y, id_: str) -> None:
        with self._gen_wlock():
            idx = self._id2idx.get(id_)
            if idx is None:
                if id_ in self._y_waiting_room:
                    raise ServiceErrorByDuplicatedId
                self._y_waiting_room[id_] = y
                listener_and_idx = None
            else:
                if not self._core.recv_y(idx, y):
                    raise ServiceErrorByDuplicatedId
                listener_and_idx = self._listener, idx
        if listener_and_idx is not None:
            listener, idx = listener_and_idx
            if listener is not None:
                listener.notify_y(idx, y)

    def register_listener(self, listener: BaseDataHistoryListener[X, Y, YP] | None) -> None:
        with self._gen_wlock():
            self._listener = listener


class DataHistoryInMemoryCore(Generic[X, Y, YP]):

    def __init__(self) -> None:
        self._items = list[DataHistoryItem[X, Y, YP]]()

    def __len__(self) -> int:
        return len(self._items)

    def __getitem__(self, idx: int) -> DataHistoryItem[X, Y, YP]:
        return self._items[idx]

    def append(self, x: X, y: Y | DataHistoryItemYNone, y_pred: ModelPredictionProcess[X, YP]) -> None:
        self._items.append(DataHistoryItem(x, y, y_pred))

    def recv_y(self, idx: int, y: Y) -> bool:
        if not 0 <= idx < len(self._items):
            raise RuntimeError
        if not isinstance(self._items[idx].y, DataHistoryItemYNone):
            return False
        self._items[idx].y = y
        return True

    def recv_y_pred(self, idx: int, y_pred: Prediction[YP]) -> bool:
        if not 0 <= idx < len(self._items):
            raise RuntimeError
        if not isinstance(self._items[idx].y_pred, ModelPredictionProcess):
            return False
        self._items[idx].y_pred = y_pred
        return True


__all__ = [
    'DataHistoryInMemory',
]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar


R = TypeVar('R')


class BaseResourcePool(Generic[R], ABC):

    @abstractmethod
    def alloc(self) -> R | None:
        pass

    @abstractmethod
    def dealloc(self, resource: R) -> None:
        pass


__all__ = [
    'BaseResourcePool',
]



from collections.abc import Iterable, Sequence
from typing import Any, Literal, overload

from musc.evaluator.a import Evaluator as EvaluatorLowLevel
from musc.evaluator.a2 import Evaluator2 as EvaluatorLowLevel2
from musc.service.concepts.a import Metric, UpdateStrategy
from musc.service.concepts.high_level.a import apply_model_adaptors
from musc.service.impl.dd_process.stats import Stats


class Evaluator:

    def __init__(
        self,
        model: Any,
        update_strategy: UpdateStrategy,
        mode: Literal['simulation'] | Literal['idle_detect'] = 'simulation',
    ) -> None:
        model_ = apply_model_adaptors(model)
        update_strategy_ = update_strategy.clone_without_state()
        match mode:
            case 'simulation':
                self._inner = EvaluatorLowLevel2(model_, update_strategy_)
            case 'idle_detect':
                self._inner = EvaluatorLowLevel(model_, update_strategy_)
            case _:
                raise ValueError

    @overload
    def evaluate(
        self,
        x_arr: Iterable[Any],
        y_arr: Iterable[Any],
        t_x_arr: Iterable[Any],
        t_y_arr: Iterable[Any],
    ) -> tuple[list[Any], Stats]:
        pass

    @overload
    def evaluate(
        self,
        x_arr: Iterable[Any],
        y_arr: Iterable[Any],
        t_x_arr: Iterable[Any],
        t_y_arr: Iterable[Any],
        metric: Metric,
    ) -> tuple[float, Stats]:
        pass

    @overload
    def evaluate(
        self,
        x_arr: Iterable[Any],
        y_arr: Iterable[Any],
        t_x_arr: Iterable[Any],
        t_y_arr: Iterable[Any],
        metric: Sequence[Metric],
    ) -> tuple[list[float], Stats]:
        pass

    def evaluate(
        self,
        x_arr: Iterable[Any],
        y_arr: Iterable[Any],
        t_x_arr: Iterable[Any],
        t_y_arr: Iterable[Any],
        metric: Metric | Sequence[Metric] | None = None,
    ) -> tuple[float | list[float] | list[Any], Stats]:

        x_arr_, y_arr_, t_x_arr_, t_y_arr_ = list(x_arr), list(y_arr), list(t_x_arr), list(t_y_arr)
        del x_arr, y_arr, t_x_arr, t_y_arr
        if not len(x_arr_) == len(y_arr_) == len(t_x_arr_) == len(t_y_arr_) != 0:
            raise ValueError

        if metric is None:
            metric_ = None
        elif isinstance(metric, Metric):
            metric_ = metric
        else:
            metric_ = list(metric)
        del metric

        y_pred_arr, stats = self._inner.evaluate(x_arr_, y_arr_, t_x_arr_, t_y_arr_)
        if metric_ is None:
            return y_pred_arr, stats
        elif isinstance(metric_, Metric):
            return self._score(y_arr_, y_pred_arr, metric_), stats
        else:
            return [self._score(y_arr_, y_pred_arr, m) for m in metric_], stats

    def _score(self, y_arr: list[Any], y_pred_arr: list[Any], metric: Metric) -> float:
        model = self._inner.get_model()
        return sum(
            metric(y=model.preprocess_y(y), y_pred=y_pred)
            for y, y_pred in zip(y_arr, y_pred_arr, strict=True)
        ) / len(y_arr)


__all__ = [
    'Evaluator',
]



from collections.abc import Callable, Generator, Iterable
from multiprocessing import Manager
from os import PathLike
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Any, TypeVar
from uuid import uuid4

import cloudpickle
from joblib import Parallel, delayed


R, T, V = TypeVar('R'), TypeVar('T'), TypeVar('V')


class _JoblibFinished:
    pass


def explain_object(obj: Any) -> str:
    if hasattr(obj, '__explain_self__') and not isinstance(obj, type):
        return obj.__explain_self__()
    if hasattr(obj, '__name__'):
        if hasattr(obj, '__module__') and obj.__module__ != '__main__':
            return f'{obj.__module__}.{obj.__name__}'
        return obj.__name__
    repr_obj = repr(obj)
    if repr_obj.startswith('<'):
        return f'{explain_object(type(obj))}(...)'
    return repr_obj


def joblib_imap_unordered(
    parallel: Parallel,
    func: Callable[[V], R],
    iter: Iterable[V],
) -> Generator[R, None, None]:
    with Manager() as multiprocessing_manager:
        queue: Queue[bytes | _JoblibFinished] = multiprocessing_manager.Queue()
        progress = parallel(map(delayed(lambda v: queue.put(cloudpickle.dumps(func(v)))), iter))
        monitor_thread = Thread(target=_joblib_progress_monitor, args=(queue, progress))
        monitor_thread.start()
        while not isinstance(value := queue.get(), _JoblibFinished):
            yield cloudpickle.loads(value)
        monitor_thread.join()


def safe_save(original_save: Callable[[str], T], path: str | PathLike) -> T:
    path = Path(path)
    tmp_path = str(uuid4()) + ''.join(path.suffixes)
    return_value = original_save(tmp_path)
    path.unlink(missing_ok=True)
    Path(tmp_path).rename(path)
    return return_value


def _joblib_progress_monitor(queue: Queue[Any], progress: Iterable[Any]) -> Any:
    for _ in progress:
        pass
    queue.put(_JoblibFinished())


__all__ = [
    'explain_object',
    'joblib_imap_unordered',
    'safe_save',
]



from __future__ import annotations

import logging
import typing
from dataclasses import dataclass
from threading import Lock, Thread
from typing import Generic, TypeVar

from readerwriterlock.rwlock import RWLockFair

from musc.service.concepts.a import BaseModel, Updator


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')


class ModelWrapper(Generic[XR, X, YR, Y, YP]):

    def __init__(
        self,
        model: BaseModel[XR, X, YR, Y, YP],
        updator: Updator[X, Y],
        daemon: bool,
        debug: bool,
    ) -> None:
        self._model = model
        self._gen = 0
        self._updator = updator
        self._daemon = daemon
        self._debug = debug
        self._model_lock = RWLockFair()
        self._update_lock = Lock()

    def gen(self) -> int:
        with self._model_lock.gen_rlock():
            gen = self._gen
        return gen

    def predict(self, x: X) -> ModelPredictionProcess[X, YP]:
        with self._model_lock.gen_rlock():
            model = self._model
            gen = self._gen
        return ModelPredictionProcess(model, gen, x, self._daemon, self._debug)

    def update_by(self, x_arr: list[X], y_arr: list[Y]) -> None:
        with self._update_lock:
            model_new = self._model.clone()
            try:
                self._updator(model_new, x_arr, y_arr)
            except Exception as e:
                if self._debug:
                    raise
                logging.error(f'The model updator raised an exception: {repr(e)}')
                with self._model_lock.gen_wlock():
                    self._gen += 1
            else:
                with self._model_lock.gen_wlock():
                    self._model = model_new
                    self._gen += 1

    def preprocess_x(self, x: XR) -> X:
        with self._model_lock.gen_rlock():
            model = self._model
        return model.preprocess_x(x)

    def preprocess_y(self, y: YR) -> Y:
        with self._model_lock.gen_rlock():
            model = self._model
        return model.preprocess_y(y)


class ModelPredictionProcess(Generic[X, YP]):

    def __init__(
        self,
        model: BaseModel[XR, X, YR, Y, YP],
        gen: int,
        x: X,
        daemon: bool,
        debug: bool,
    ) -> None:

        self._model = model
        self._gen = gen
        self._x = x
        self._debug = debug

        self._success = None
        self._y_pred = None

        self._thread = Thread(target=self._thread_target, daemon=daemon)
        self._thread.start()

    def _thread_target(self) -> None:
        try:
            y_pred = self._model.predict(self._x)
        except Exception as e:
            if self._debug:
                raise
            self._success = False
            self._y_pred = e
        else:
            self._success = True
            self._y_pred = y_pred

    def gen(self) -> int:
        return self._gen

    def join(self) -> Prediction[YP]:
        self._thread.join()
        assert isinstance(self._success, bool)
        return Prediction(self._success, typing.cast(YP | Exception, self._y_pred), self._gen)


@dataclass
class Prediction(Generic[YP]):
    success: bool
    y_pred: YP | Exception
    gen: int


__all__ = [
    'ModelWrapper',
    'ModelPredictionProcess',
    'Prediction',
]



from __future__ import annotations

import multiprocessing
from multiprocessing import Manager, Pipe
from multiprocessing.connection import _ConnectionBase
from queue import Queue
from typing import Any, Generic, TypeVar

from musc.service.concepts.a import BaseModel
from musc.service.concepts.high_level.a import ModelGuard


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')


QueueToDaemon = Queue[_ConnectionBase | None]


class ModelCloneHelper(Generic[XR, X, YR, Y, YP]):

    def __init__(self, model: BaseModel[XR, X, YR, Y, YP]) -> None:
        self.__multiprocessing_manager = Manager()
        self.__queue_to_daemon: QueueToDaemon = self.__multiprocessing_manager.Queue()
        # self._process cannot be sent to joblib process pool.
        # Use 'spawn' here. See https://github.com/pytorch/pytorch/issues/122186.
        self.__process = multiprocessing.get_context('spawn').Process(
            target=process_target,
            args=(self.__queue_to_daemon, model),
            daemon=True,
        )
        self.__process.start()

    def portable(self) -> ModelCloneHelperPortable[XR, X, YR, Y, YP]:
        return ModelCloneHelperPortable(self.__queue_to_daemon)

    def stop(self) -> None:
        self.__queue_to_daemon.put(None)
        self.__process.join()
        self.__multiprocessing_manager.shutdown()


class ModelCloneHelperPortable(Generic[XR, X, YR, Y, YP], ModelGuard[XR, X, YR, Y, YP]):

    def __init__(self, queue_to_daemon: QueueToDaemon) -> None:
        self.__queue_to_daemon = queue_to_daemon

    def __get(self) -> BaseModel[XR, X, YR, Y, YP]:
        resp_receiver, resp_sender = Pipe(duplex=False)
        self.__queue_to_daemon.put(resp_sender)
        return resp_receiver.recv()

    def clone(self) -> BaseModel[XR, X, YR, Y, YP]:
        return self.__get().clone()

    def clone_with_resource(self, resource: Any) -> BaseModel[XR, X, YR, Y, YP]:
        return self.__get().clone_with_resource(resource)


def process_target(queue: QueueToDaemon, model: BaseModel[XR, X, YR, Y, YP]) -> None:
    while True:
        match queue.get():
            case None:
                break
            case resp_sender:
                resp_sender.send(model)


__all__ = [
    'ModelCloneHelper',
]



from __future__ import annotations

import time
import typing
from multiprocessing import Manager, Pipe, Process
from multiprocessing.connection import _ConnectionBase
from queue import Queue
from typing import Generic, Literal, TypeVar

from musc.hpo.res_pool.base import BaseResourcePool


R = TypeVar('R')


QueueToDaemon = Queue[tuple[Literal['alloc'], _ConnectionBase] | tuple[Literal['dealloc'], R] | None]


class GlobalResourcePool(Generic[R]):

    def __init__(self, inner: BaseResourcePool[R]) -> None:
        self._multiprocessing_manager = Manager()
        self._queue_to_daemon: QueueToDaemon[R] = self._multiprocessing_manager.Queue()
        # self._process cannot be sent to joblib process pool.
        self._process = Process(
            target=process_target,
            args=(self._queue_to_daemon, inner),
            daemon=True,
        )
        self._process.start()

    def portable(self) -> GlobalResourcePoolPortable[R]:
        return GlobalResourcePoolPortable(self._queue_to_daemon)

    def stop(self) -> None:
        self._queue_to_daemon.put(None)
        self._process.join()
        self._multiprocessing_manager.shutdown()


class GlobalResourcePoolPortable(Generic[R]):

    def __init__(self, queue_to_daemon: QueueToDaemon[R]) -> None:
        self._queue_to_daemon = queue_to_daemon

    def alloc(self) -> ResourceContext[R]:
        return ResourceContext(self, self.alloc_manual())

    def alloc_manual(self) -> R:
        while True:
            resp_receiver, resp_sender = Pipe(duplex=False)
            self._queue_to_daemon.put(('alloc', resp_sender))
            if (resource := typing.cast(R | None, resp_receiver.recv())) is not None:
                return resource
            time.sleep(1.0)

    def dealloc_manual(self, resource: R) -> None:
        self._queue_to_daemon.put(('dealloc', resource))


class ResourceContext(Generic[R]):

    def __init__(self, pool: GlobalResourcePoolPortable[R], resource: R) -> None:
        self._pool = pool
        self._resource = resource

    def __enter__(self) -> R:
        return self._resource

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        del exc_type, exc_value, traceback
        self._pool.dealloc_manual(self._resource)


def process_target(queue: QueueToDaemon[R], pool: BaseResourcePool[R]) -> None:
    while True:
        match queue.get():
            case ('alloc', resp_sender):
                resp_sender.send(pool.alloc())
            case ('dealloc', resource):
                pool.dealloc(resource)
            case None:
                break


__all__ = [
    'GlobalResourcePool',
]



from __future__ import annotations

import random
from typing import Any

from musc.service.concepts.a import BaseDriftDetectorWithNcl, DriftDetected


class DriftDetectorNone(BaseDriftDetectorWithNcl):

    def step(self, x: Any, y: Any, y_pred: Any) -> DriftDetected | None:
        del x, y, y_pred
        return None

    def clone_without_state(self) -> DriftDetectorNone:
        return DriftDetectorNone()


class DriftDetectorPeriodicallyUpdate(BaseDriftDetectorWithNcl):

    def __init__(
        self,
        period: int,
        ncl: int | None = None,
        probability: float | None = None,
    ) -> None:
        assert period >= 1
        assert probability is None or 0.0 <= probability <= 1.0
        self._period = period
        self._ncl = ncl if ncl is not None else period
        self._cnt = 0
        self._probability = probability

    def step(self, x: Any, y: Any, y_pred: Any) -> DriftDetected | None:
        del x, y, y_pred
        self._cnt += 1
        if self._cnt % self._period == 0:
            if self._probability is not None and random.random() >= self._probability:
                return None
            if self._ncl <= -1:
                return DriftDetected(self._cnt)
            else:
                return DriftDetected(min(self._ncl, self._cnt))
        else:
            return None

    def clone_without_state(self) -> DriftDetectorPeriodicallyUpdate:
        return DriftDetectorPeriodicallyUpdate(self._period, self._ncl, self._probability)


__all__ = [
    'DriftDetectorNone',
    'DriftDetectorPeriodicallyUpdate',
]



from __future__ import annotations

import queue
import typing
from queue import Queue
from threading import Lock
from typing import Generic, Literal, TypeVar, overload

from musc.evaluator.active_count import BaseActiveCount, NullActiveCount
from musc.service.impl.model_wrapper import ModelPredictionProcess, ModelWrapper, Prediction
from musc.service.impl.data_history.base import (
    BaseDataHistory,
    BaseDataHistoryListener,
    DataHistoryItemYNone,
)


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')
X_, Y_, YP_ = TypeVar('X_'), TypeVar('Y_'), TypeVar('YP_')


class DataHistoryWalker(Generic[X, Y, YP]):

    class Listener(Generic[X_, Y_, YP_], BaseDataHistoryListener[X_, Y_, YP_]):

        def __init__(self, active_count: BaseActiveCount) -> None:
            self._x_queues = dict[int, Queue[tuple[X_, ModelPredictionProcess[X_, YP_]] | Never]]()
            self._y_queues = dict[int, Queue[Y_ | Never]]()
            self._stopped = False
            self._active_count = active_count
            self._lock = Lock()

        def request_x(self, idx: int) -> Queue[tuple[X_, ModelPredictionProcess[X_, YP_]] | Never]:
            with self._lock:
                if idx in self._x_queues:
                    raise RuntimeError
                queue = Queue[tuple[X_, ModelPredictionProcess[X_, YP_]] | Never]()
                if self._stopped:
                    self._active_count.inc()
                    queue.put(Never())
                else:
                    self._x_queues[idx] = queue
                return queue

        def request_y(self, idx: int) -> Queue[Y_ | Never]:
            with self._lock:
                if idx in self._y_queues:
                    raise RuntimeError
                queue = Queue[Y_ | Never]()
                if self._stopped:
                    self._active_count.inc()
                    queue.put(Never())
                else:
                    self._y_queues[idx] = queue
                return queue

        def notify_x(self, idx: int, x: X_, y_pred: ModelPredictionProcess[X_, YP_]) -> None:
            with self._lock:
                if idx in self._x_queues:
                    self._active_count.inc()
                    self._x_queues.pop(idx).put((x, y_pred))

        def notify_y(self, idx: int, y: Y_) -> None:
            with self._lock:
                if idx in self._y_queues:
                    self._active_count.inc()
                    self._y_queues.pop(idx).put(y)

        def stop(self) -> None:
            with self._lock:
                self._stopped = True
                for idx in list(self._x_queues.keys()):
                    self._active_count.inc()
                    self._x_queues.pop(idx).put(Never())
                for idx in list(self._y_queues.keys()):
                    self._active_count.inc()
                    self._y_queues.pop(idx).put(Never())

    def __init__(
        self,
        data_history: BaseDataHistory[X, Y, YP],
        model: ModelWrapper[XR, X, YR, Y, YP],
        wait_y_timeout: float | None,
        active_count: BaseActiveCount,
    ) -> None:
        assert isinstance(active_count, NullActiveCount) or wait_y_timeout is None
        self._data_history = data_history
        self._model = model
        self._wait_y_timeout = wait_y_timeout
        self._active_count = active_count
        self._listener = DataHistoryWalker.Listener[X, Y, YP](active_count)
        self._data_history.register_listener(self._listener)
        self._cnt = 0
        self._err_indices = set[int]()

    def _wait_x(self) -> tuple[X, ModelPredictionProcess[X, YP] | Prediction[YP]] | Never:
        with self._data_history.gen_rlock():
            if self._cnt < len(self._data_history):
                cur_item = self._data_history[self._cnt]
                return cur_item.x, cur_item.y_pred
            queue = self._listener.request_x(self._cnt)
        self._active_count.dec()
        return queue.get()

    def _wait_y_after_x(self) -> Y | DataHistoryItemYNone | Never:
        with self._data_history.gen_rlock():
            if not isinstance(y := self._data_history[self._cnt].y, DataHistoryItemYNone):
                return y
            queue_ = self._listener.request_y(self._cnt)
        self._active_count.dec()
        try:
            return queue_.get(timeout=self._wait_y_timeout)
        except queue.Empty:
            return DataHistoryItemYNone()

    @overload
    def step(self) -> tuple[X, Y, YP] | Never:
        pass

    @overload
    def step(self, require_y_pred: Literal[True]) -> tuple[X, Y, YP] | Never:
        pass

    @overload
    def step(self, require_y_pred: Literal[False]) -> tuple[X, Y, None] | Never:
        pass

    def step(self, require_y_pred: bool = True) -> tuple[X, Y, YP | None] | Never:
        while True:
            wait_result = self._wait_x()
            if isinstance(wait_result, Never):
                return wait_result
            x, y_pred = wait_result
            if require_y_pred:
                if isinstance(y_pred, ModelPredictionProcess):
                    if y_pred.gen() != self._model.gen():
                        y_pred = self._model.predict(x).join()
                    else:
                        y_pred = y_pred.join()
                elif isinstance(y_pred, Prediction):
                    if y_pred.gen != self._model.gen():
                        y_pred = self._model.predict(x).join()
                else:
                    assert False
                if not y_pred.success:
                    self._err_indices.add(self._cnt)
                    self._cnt += 1
                    continue
                y_pred = typing.cast(YP, y_pred.y_pred)
            else:
                y_pred = None
            wait_result = self._wait_y_after_x()
            if isinstance(wait_result, Never):
                return wait_result
            if isinstance(wait_result, DataHistoryItemYNone):
                self._err_indices.add(self._cnt)
                self._cnt += 1
                continue
            y = typing.cast(Y, wait_result)
            self._cnt += 1
            return x, y, y_pred

    def cnt(self) -> int:
        return self._cnt

    def err_indices(self) -> set[int]:
        return self._err_indices

    def stop_listening(self) -> None:
        self._listener.stop()


class Never:
    pass


__all__ = [
    'DataHistoryWalker',
    'Never',
]



from typing import Any

import river.drift

from musc.hpo.distr.a import BaseDistribution
from musc.hpo.distr.high_level import into_distr


class ADWIN(river.drift.ADWIN):

    @staticmethod
    def kwargs_distr_prelude() -> dict[str, BaseDistribution[Any]]:
        return {
            'delta': into_distr([
                0.0001, 0.0002, 0.0005, 0.001, 0.002, 0.005, 0.01, 0.02, 0.05,
                0.1, 0.2, 0.5, 0.8, 0.9, 0.95, 0.98, 0.99, 0.995, 0.998, 0.999,
            ]),
            'clock': into_distr([1]),
            'max_buckets': into_distr([4, 6, 8, 12, 16]),
            'min_window_length': into_distr([2, 3, 4, 6, 8, 12]),
            'grace_period': into_distr([2, 3, 4, 6, 8, 12, 16, 24]),
        }


class KSWIN(river.drift.KSWIN):

    @staticmethod
    def kwargs_distr_prelude() -> dict[str, BaseDistribution[Any]]:
        return {
            'alpha': into_distr([
                0.00001, 0.00002, 0.00005, 0.0001, 0.0002, 0.0005,
                0.001, 0.002, 0.005, 0.01, 0.02, 0.05, 0.1, 0.2, 0.5, 0.8, 0.9,
            ]),
        }


class PageHinkley(river.drift.PageHinkley):

    @staticmethod
    def kwargs_distr_prelude() -> dict[str, BaseDistribution[Any]]:
        return {
            'min_instances': into_distr([2, 3, 4, 6, 8, 12, 16, 24, 32, 48, 64]),
            'delta': into_distr([0.0005, 0.001, 0.002, 0.005, 0.01, 0.02, 0.05]),
            'threshold': into_distr([
                0.0001, 0.0002, 0.0005, 0.001, 0.002, 0.005, 0.01, 0.02, 0.05,
                0.1, 0.2, 0.5, 1.0, 2.0, 5.0, 10.0, 20.0, 50.0,
            ]),
        }


def kswin_distr_prelude() -> BaseDistribution[KSWIN]:
    return into_distr([
        {
            'type': KSWIN,
            **KSWIN.kwargs_distr_prelude(),
            'window_size': [stat_size * r for r in [2, 3, 4, 6]],
            'stat_size': stat_size,
        }
        for stat_size in [2, 3, 4, 6, 8, 12, 16, 24, 32, 48, 64, 96, 128]
    ])


__all__ = [
    'ADWIN',
    'KSWIN',
    'PageHinkley',
    'kswin_distr_prelude',
]



from dataclasses import dataclass


class ServiceErrorByDuplicatedId(Exception):
    pass


@dataclass
class ServiceErrorFromModelPrediction(Exception):
    inner: BaseException


__all__ = [
    'ServiceErrorByDuplicatedId',
    'ServiceErrorFromModelPrediction',
]



from collections import Counter
from collections.abc import Sequence

import torch

from musc.hpo.res_pool.base import BaseResourcePool


class BaseTorchDevicePool(BaseResourcePool[torch.device]):

    def __init__(self, indices: Sequence[int], type_: str) -> None:
        self._index_counter = Counter(indices)
        self._type = type_

    def alloc(self) -> torch.device | None:
        if (index := next(iter(self._index_counter.keys()), None)) is None:
            return None
        self._index_counter.subtract([index])
        self._index_counter = +self._index_counter
        return torch.device(self._type, index)

    def dealloc(self, resource: torch.device) -> None:
        if resource.type != self._type:
            raise ValueError
        self._index_counter.update([resource.index])


class TorchDevicePool(BaseTorchDevicePool):

    def __init__(self, indices: Sequence[int], repeat: int = 1, type_: str = 'cuda') -> None:
        super().__init__(list(indices) * repeat, type_)


__all__ = [
    'TorchDevicePool',
]



import copy
import time
import typing
from collections.abc import Callable
from dataclasses import dataclass
from threading import Thread
from typing import Generic, TypeVar

from musc.evaluator.active_count import BaseActiveCount
from musc.service.concepts.a import UpdateStrategy
from musc.service.impl.model_wrapper import ModelWrapper
from musc.service.impl.data_history.base import BaseDataHistory
from musc.service.impl.dd_process.dh_walker import DataHistoryWalker, Never
from musc.service.impl.dd_process.stats import ModelUpdateRecord, Stats, StatsWithLock


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')
X_, Y_ = TypeVar('X_'), TypeVar('Y_')
T = TypeVar('T')


class DriftDetectorProcess(Generic[X, Y, YP]):

    @dataclass
    class ModeDetecting:
        pass

    @dataclass
    class ModeCollecting(Generic[X_, Y_]):
        x_arr: list[X_]
        y_arr: list[Y_]

    def __init__(
        self,
        data_history: BaseDataHistory[X, Y, YP],
        update_strategy: UpdateStrategy[X, Y, YP],
        model: ModelWrapper[XR, X, YR, Y, YP],
        wait_y_timeout: float | None,
        daemon: bool,
        active_count: BaseActiveCount,
    ) -> None:

        self._data_history = data_history
        self._update_strategy = update_strategy
        self._model = model

        self._data_history_walker = DataHistoryWalker(
            self._data_history,
            self._model,
            wait_y_timeout,
            active_count,
        )
        self._mode = DriftDetectorProcess.ModeDetecting()
        self._last_drift_point = None
        self._last_new_concept_range_start = None
        self._stats = StatsWithLock()

        self._active_count = active_count
        self._thread = Thread(target=self._thread_target, daemon=daemon)
        self._active_count.inc()
        self._thread.start()

    def _thread_target(self) -> None:

        while True:

            if isinstance(self._mode, DriftDetectorProcess.ModeDetecting):

                step_result = self._data_history_walker.step()
                if isinstance(step_result, Never):
                    break
                x, y, y_pred = step_result

                if (ncl := self._update_strategy.drift_detector.step(x, y, y_pred)) is None:
                    continue
                if self._update_strategy.use_ncl:
                    ncl = ncl.ncl
                    if ncl is None or ncl < 0:
                        raise RuntimeError
                else:
                    ncl = 0

                x_arr, y_arr = list[X](), list[Y]()
                self._last_new_concept_range_start = self._data_history_walker.cnt()
                self._last_drift_point = self._data_history_walker.cnt()
                with self._data_history.gen_rlock():
                    while True:
                        if len(x_arr) == ncl:
                            break
                        if self._last_new_concept_range_start == 0:
                            raise RuntimeError
                        if self._last_new_concept_range_start - 1 \
                                not in self._data_history_walker.err_indices():
                            cur_item = self._data_history[self._last_new_concept_range_start - 1]
                            x_arr.append(cur_item.x)
                            y_arr.append(typing.cast(Y, cur_item.y))
                        self._last_new_concept_range_start -= 1
                x_arr.reverse()
                y_arr.reverse()

                if len(x_arr) < self._update_strategy.data_amount_required:
                    self._mode = DriftDetectorProcess.ModeCollecting(x_arr, y_arr)
                else:
                    self._update(x_arr, y_arr)

            elif isinstance(self._mode, DriftDetectorProcess.ModeCollecting):

                step_result = self._data_history_walker.step(require_y_pred=False)
                if isinstance(step_result, Never):
                    break
                x, y, _ = step_result

                self._mode.x_arr.append(x)
                self._mode.y_arr.append(y)

                if len(self._mode.x_arr) == self._update_strategy.data_amount_required:
                    self._update(self._mode.x_arr, self._mode.y_arr)
                    self._mode = DriftDetectorProcess.ModeDetecting()

            else:

                assert False

        self._active_count.dec()

    def _update(self, x_arr: list[X], y_arr: list[Y]) -> None:
        assert self._last_drift_point is not None
        assert self._last_new_concept_range_start is not None
        start_time = time.time()
        self._model.update_by(x_arr, y_arr)
        time_spent = time.time() - start_time
        with self._stats.gen_wlock():
            self._stats.inner().model_update_records.append(
                ModelUpdateRecord(
                    self._last_drift_point,
                    (self._last_new_concept_range_start, self._data_history_walker.cnt()),
                    time_spent,
                )
            )

    def stats(self, fn: Callable[[Stats], T] = lambda s: s) -> T:
        with self._stats.gen_rlock():
            return copy.deepcopy(fn(self._stats.inner()))

    def stop_listening(self) -> None:
        self._data_history_walker.stop_listening()

    def join(self) -> None:
        return self._thread.join()


__all__ = [
    'DriftDetectorProcess',
]



from __future__ import annotations

import operator
import time
from collections.abc import Callable
from typing import Generic, TypeVar

from musc.service.concepts.a import BaseModel, UpdateStrategy
from musc.service.impl.dd_process.a import DriftDetectorProcess
from musc.service.impl.dd_process.stats import ModelUpdateRecord, Stats


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')
XR_, X_, YR_, Y_, YP_ = TypeVar('XR_'), TypeVar('X_'), TypeVar('YR_'), TypeVar('Y_'), TypeVar('YP_')
V = TypeVar('V')


class Evaluator2(Generic[XR, X, YR, Y, YP]):

    def __init__(
        self,
        model: BaseModel[XR, X, YR, Y, YP],
        update_strategy: UpdateStrategy[X, Y, YP],
    ) -> None:

        self.__model_history = [(model, float('-inf'), int(0))]
        self.__update_strategy = update_strategy

        self.__evaluated = False

        self.__y_pred_arr = list[tuple[YP, float, int] | None]()

        self._current_time = float('-inf')
        self.__mode = DriftDetectorProcess.ModeDetecting()
        self.__last_drift_point = None
        self.__last_new_concept_range_start = None
        self.__stats = Stats([])

    def get_model(self) -> BaseModel[XR, X, YR, Y, YP]:
        return self.__model_history[-1][0]

    def evaluate(
        self,
        x_arr: list[XR],
        y_arr: list[YR],
        t_x_arr: list[float],
        t_y_arr: list[float],
    ) -> tuple[list[YP], Stats]:

        if not len(x_arr) == len(y_arr) == len(t_x_arr) == len(t_y_arr):
            raise ValueError

        if self.__evaluated:
            raise RuntimeError
        self.__evaluated = True

        x_arr_preprocessed = list[X]()
        t_x_arr_preprocessed = list[float]()
        for x, t_x in zip(x_arr, t_x_arr, strict=True):
            x_p, t_p = self.timeit(lambda: self.__model_history[0][0].preprocess_x(x))
            x_arr_preprocessed.append(x_p)
            t_x_arr_preprocessed.append(t_x + t_p)

        y_arr_preprocessed = list[Y]()
        t_y_arr_preprocessed = list[float]()
        for y, t_y in zip(y_arr, t_y_arr, strict=True):
            y_p, t_p = self.timeit(lambda: self.__model_history[0][0].preprocess_y(y))
            y_arr_preprocessed.append(y_p)
            t_y_arr_preprocessed.append(t_y + t_p)

        return self.__evaluate(
            x_arr_preprocessed,
            y_arr_preprocessed,
            t_x_arr_preprocessed,
            t_y_arr_preprocessed,
        )

    def __evaluate(
        self,
        x_arr: list[X],
        y_arr: list[Y],
        t_x_arr: list[float],
        t_y_arr: list[float],
    ) -> tuple[list[YP], Stats]:

        data = zip(x_arr, y_arr, t_x_arr, t_y_arr, range(len(x_arr)), strict=True)
        data = sorted(data, key=operator.itemgetter(2))
        del x_arr, y_arr, t_x_arr, t_y_arr

        self.__y_pred_arr.extend([None] * len(data))
        data_cursor = 0

        for cnt, (x, y, t_x, t_y, id_) in enumerate(data, start=1):
            while data_cursor < len(data):
                x_, _, t_x_, _, id__ = data[data_cursor]
                if t_x_ > max(t_x, t_y):
                    break
                self.__prune_model_history_by_t_x(t_x_)
                y_pred_, t_p_ = self.timeit(lambda: self.__model_history[0][0].predict(x_))
                self.__y_pred_arr[id__] = (y_pred_, t_p_, self.__model_history[0][2])
                data_cursor += 1
            if isinstance(self.__mode, DriftDetectorProcess.ModeDetecting):
                y_pred = self.__determine_t_dhw_step(x, t_x, t_y, id_)
                with self.Timer(self) as timer:
                    if (ncl := self.__update_strategy.drift_detector.step(x, y, y_pred)) is None:
                        continue
                    if self.__update_strategy.use_ncl:
                        ncl = ncl.ncl
                        if ncl is None or ncl < 0:
                            raise RuntimeError
                    else:
                        ncl = 0
                    data_ = data[cnt-ncl:cnt]
                    x_arr_, y_arr_ = [d[0] for d in data_], [d[1] for d in data_]
                    self.__last_new_concept_range_start, self.__last_drift_point = cnt - ncl, cnt
                    if len(x_arr_) < self.__update_strategy.data_amount_required:
                        self.__mode = DriftDetectorProcess.ModeCollecting(x_arr_, y_arr_)
                    else:
                        self.__update(x_arr_, y_arr_, cnt, timer)
            elif isinstance(self.__mode, DriftDetectorProcess.ModeCollecting):
                self.__determine_t_dhw_step_woyp(t_x, t_y)
                with self.Timer(self) as timer:
                    self.__mode.x_arr.append(x)
                    self.__mode.y_arr.append(y)
                    if len(self.__mode.x_arr) == self.__update_strategy.data_amount_required:
                        self.__update(self.__mode.x_arr, self.__mode.y_arr, cnt, timer)
                        self.__mode = DriftDetectorProcess.ModeDetecting()
            else:
                assert False

        y_pred_arr = list[YP]()
        for item in self.__y_pred_arr:
            assert item is not None
            y_pred_arr.append(item[0])
        return y_pred_arr, self.__stats

    def __determine_t_dhw_step(self, x: X, t_x: float, t_y: float, id_: int) -> YP:
        assert (y_pred_item := self.__y_pred_arr[id_]) is not None
        y_pred, t_p, gen = y_pred_item
        if gen == self.__model_history[-1][2]:
            self._current_time = max(self._current_time, t_x + t_p, t_y)
        else:
            y_pred, t_p = self.timeit(lambda: self.__model_history[-1][0].predict(x))
            self._current_time = max(max(self._current_time, t_x) + t_p, t_y)
        return y_pred

    def __determine_t_dhw_step_woyp(self, t_x: float, t_y: float) -> None:
        self._current_time = max(self._current_time, t_x, t_y)

    def __prune_model_history_by_t_x(self, t_x: float) -> None:
        for i in reversed(range(len(self.__model_history))):
            if self.__model_history[i][1] <= t_x:
                del self.__model_history[:i]
                return
        assert False

    def __update(
        self,
        x_arr: list[X],
        y_arr: list[Y],
        cnt: int,
        timer: Timer[XR, X, YR, Y, YP],
    ) -> None:
        assert self.__last_drift_point is not None
        assert self.__last_new_concept_range_start is not None
        _, time_spent = self.timeit(lambda: self.__update_by(x_arr, y_arr, timer))
        self.__stats.model_update_records.append(ModelUpdateRecord(
            self.__last_drift_point,
            (self.__last_new_concept_range_start, cnt),
            time_spent,
        ))

    def __update_by(self, x_arr: list[X], y_arr: list[Y], timer: Timer[XR, X, YR, Y, YP]) -> None:
        model_new = self.__model_history[-1][0].clone()
        self.__update_strategy.updator(model_new, x_arr, y_arr)
        self.__model_history.append((model_new, timer.get(), self.__model_history[-1][2] + 1))

    @staticmethod
    def timeit(callable_: Callable[[], V]) -> tuple[V, float]:
        start_time = time.time()
        return callable_(), time.time() - start_time

    class Timer(Generic[XR_, X_, YR_, Y_, YP_]):

        def __init__(self, evaluator: Evaluator2[XR_, X_, YR_, Y_, YP_]) -> None:
            self.__evaluator = evaluator
            self.__start_time = None

        def get(self) -> float:
            if self.__start_time is None:
                raise RuntimeError
            return self.__evaluator._current_time + (time.time() - self.__start_time)

        def __enter__(self) -> Evaluator2.Timer[XR_, X_, YR_, Y_, YP_]:
            self.__start_time = time.time()
            return self

        def __exit__(self, exc_type, exc_value, traceback) -> None:
            del exc_type, exc_value, traceback
            self.__evaluator._current_time = self.get()
            self.__start_time = None


__all__ = [
    'Evaluator2',
]



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Generic, Literal, TypeVar, overload

import numpy as np

from musc.commons import explain_object


XR, X, YR, Y, YP = TypeVar('XR'), TypeVar('X'), TypeVar('YR'), TypeVar('Y'), TypeVar('YP')


class BaseModel(Generic[XR, X, YR, Y, YP], ABC):

    @abstractmethod
    def predict(self, x: X) -> YP:
        pass

    def inner(self) -> Any:
        return self

    def clone(self) -> BaseModel[XR, X, YR, Y, YP]:
        return deepcopy(self)

    def clone_with_resource(self, resource: Any) -> BaseModel[XR, X, YR, Y, YP]:
        del resource
        return self.clone()

    def resource_current(self) -> Any:
        return None

    def resource_cpu(self) -> Any:
        return None

    def preprocess_x(self, x: XR) -> X:
        # Implementors should ensure that the behavior of
        # this method will not be changed by model updator.
        return x  # type: ignore

    def preprocess_y(self, y: YR) -> Y:
        # Implementors should ensure that the behavior of
        # this method will not be changed by model updator.
        return y  # type: ignore

    def y_pred_to_json_like(self, y_pred: YP) -> Any:
        del y_pred
        raise NotImplementedError


@dataclass
class UpdateStrategy(Generic[X, Y, YP]):

    drift_detector: BaseDriftDetector[X, Y, YP]
    updator: Updator[X, Y]
    data_amount_required: int
    use_ncl: bool

    def __post_init__(self) -> None:
        if self.data_amount_required < 0:
            raise TypeError
        if self.use_ncl and not isinstance(self.drift_detector, BaseDriftDetectorWithNcl):
            raise TypeError

    def clone_without_state(self) -> UpdateStrategy[X, Y, YP]:
        return UpdateStrategy(
            self.drift_detector.clone_without_state(),
            self.updator,
            self.data_amount_required,
            self.use_ncl,
        )


class BaseDriftDetector(Generic[X, Y, YP], ABC):

    @abstractmethod
    def step(self, x: X, y: Y, y_pred: YP) -> DriftDetected | None:
        pass

    @abstractmethod
    def clone_without_state(self) -> BaseDriftDetector[X, Y, YP]:
        pass


class BaseDriftDetectorWithNcl(Generic[X, Y, YP], BaseDriftDetector[X, Y, YP]):

    @abstractmethod
    def clone_without_state(self) -> BaseDriftDetectorWithNcl[X, Y, YP]:
        pass


class Metric(Generic[Y, YP]):

    @overload
    def __init__(self, callable: Callable[[Y, YP], Any], pred_first: Literal[False]) -> None:
        pass

    @overload
    def __init__(self, callable: Callable[[YP, Y], Any], pred_first: Literal[True]) -> None:
        pass

    def __init__(self, callable: Callable[[Any, Any], Any], pred_first: bool) -> None:
        self._callable = callable
        self._pred_first = pred_first
        try:
            from torch import Tensor
            self._torch_tensor_class = Tensor
        except ImportError:
            self._torch_tensor_class = None

    def __call__(self, *, y: Y, y_pred: YP) -> int | float:
        if self._pred_first:
            metric_value = self._callable(y_pred, y)
        else:
            metric_value = self._callable(y, y_pred)
        if (
            self._torch_tensor_class is not None
            and isinstance(metric_value, self._torch_tensor_class)
        ):
            return metric_value.tolist()  # type: ignore
        if isinstance(metric_value, np.ndarray):
            return metric_value.tolist()
        return metric_value

    def __explain_self__(self) -> str:
        self_class_explained = explain_object(__class__)
        callable_explained = explain_object(self._callable)
        pred_first_explained = explain_object(self._pred_first)
        return f'{self_class_explained}(callable={callable_explained}, pred_first={pred_first_explained})'


@dataclass
class DriftDetected:
    ncl: int | None


# TODO: hide `BaseModel` here
Updator = Callable[[BaseModel, list[X], list[Y]], None]


__all__ = [
    'BaseModel',
    'UpdateStrategy',
    'BaseDriftDetector',
    'BaseDriftDetectorWithNcl',
    'Metric',
    'DriftDetected',
    'Updator',
]



import operator
from collections.abc import Iterable
from multiprocessing.pool import ThreadPool
from threading import Lock
from typing import Any

from musc.evaluator.active_count import ActiveCount
from musc.service.a import Service
from musc.service.concepts.a import BaseModel, UpdateStrategy
from musc.service.impl.dd_process.stats import Stats


class Evaluator:

    def __init__(
        self,
        model: BaseModel,
        update_strategy: UpdateStrategy,
        debug: bool = False,
    ) -> None:
        self._active_count, self._idle_event = ActiveCount.create()
        self._service = Service(
            model,
            update_strategy,
            daemon=False,
            debug=debug,
            active_count=self._active_count,
        )
        self._thread_pool = ThreadPool()
        self._evaluated = [False, Lock()]

    def get_model(self) -> BaseModel:
        with self._service._model._model_lock.gen_rlock():
            return self._service._model._model

    def evaluate(
        self,
        x_arr: Iterable[Any],
        y_arr: Iterable[Any],
        t_x_arr: Iterable[Any],
        t_y_arr: Iterable[Any],
    ) -> tuple[list[Any], Stats]:
        with self._evaluated[1]:
            if self._evaluated[0]:
                raise RuntimeError
            self._evaluated[0] = True
        events = []
        for i, (x, t_x) in enumerate(zip(x_arr, t_x_arr)):
            events.append((t_x, 'x', x, i))
        for i, (y, t_y) in enumerate(zip(y_arr, t_y_arr)):
            events.append((t_y, 'y', y, i))
        events.sort(key=operator.itemgetter(0))
        return self._evaluate(events)

    def _evaluate(
        self,
        events: list[tuple[float, str, Any, int]],
    ) -> tuple[list[Any], Stats]:
        last_t = None
        results = [None] * (len(events) // 2)
        results_lock = Lock()
        for t, ty, v, i in events:
            if last_t is None:
                self._idle_event.wait()
            else:
                self._idle_event.wait(t - last_t)
            last_t = t
            if ty == 'x':
                self._active_count.inc()
                self._thread_pool.apply_async(self._recv_x, (v, i, results, results_lock))
            elif ty == 'y':
                self._active_count.inc()
                self._thread_pool.apply_async(self._recv_y, (v, i))
            else:
                assert False
        self._service.stop_listening()
        self._service.join()
        self._thread_pool.close()
        self._thread_pool.join()
        return results, self._service.stats()

    def _recv_x(
        self,
        x: Any,
        id_: int,
        results: list[Any | None],
        result_lock: Lock,
    ) -> None:
        result = self._service.recv_x(x, str(id_))
        self._active_count.dec()
        with result_lock:
            results[id_] = result

    def _recv_y(self, y: Any, id_: int) -> None:
        self._service.recv_y(y, str(id_))
        self._active_count.dec()


__all__ = [
    'Evaluator',
]



from __future__ import annotations

import functools
import typing
from abc import ABC, abstractmethod
from collections.abc import Callable, Hashable, Iterable
from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from joblib import Parallel
from readerwriterlock.rwlock import RWLockFair
from sortedcontainers import SortedList as SortedListBase

from musc.commons import joblib_imap_unordered
from musc.hpo.distr.a import BaseDistribution


V, S = TypeVar('V'), TypeVar('S')
V_, S_ = TypeVar('V_'), TypeVar('S_')
T = TypeVar('T')


class BaseRandomizedSearch(Generic[V, S], ABC):

    class Portable(Generic[V_, S_]):

        def __init__(
            self,
            distr: BaseDistribution[V_],
            score_fn: Callable[[V_], S_],
            sample_callback: Callable[[Sample[V_, None]], None] | None = None,
        ) -> None:

            self._distr = distr
            self._score_fn = score_fn
            self._sample_callback = sample_callback

        def sample_by_trace(self, trace: Hashable) -> Sample[V_, S_]:
            value = self._distr.sample_by_trace(trace)
            if self._sample_callback is not None:
                self._sample_callback(Sample(self._distr, trace, value, None))
            score = self._score_fn(value)
            return Sample(self._distr, trace, value, score)

    def __init__(
        self,
        distr: BaseDistribution[V],
        score_fn: Callable[[V], S],
        greater_score_fn: Callable[[S, S], bool] | None,
        weak_greater_score_fn: Callable[[S, S], bool] | None = None,
        top_k_kept: int = 0,
        n_jobs: int | None = None,
        sample_callback: Callable[[Sample[V, None]], None] | None = None,
        survivors_callback: Callable[[Sample[V, S], list[Sample[V, S]]], bool] | None = None,
        outsider_callback: Callable[[Sample[V, S], bool], bool] | None = None,
        top_k_callback: Callable[[Sample[V, S], list[Sample[V, S]]], None] | None = None,
    ) -> None:

        if top_k_kept >= 1 and weak_greater_score_fn is None:
            raise ValueError

        self._portable = BaseRandomizedSearch.Portable(distr, score_fn, sample_callback)
        self._greater_score_fn = greater_score_fn
        self._weak_greater_score_fn = weak_greater_score_fn
        self._n_jobs = n_jobs
        self._survivors_callback = survivors_callback
        self._outsider_callback = outsider_callback
        self._top_k_callback = top_k_callback

        self._survivors = list[Sample[V, S]]()
        self._top_k = SortedList(
            key=functools.cmp_to_key(self._sample_weak_cmp),
            capacity=top_k_kept,
        )

        # self._lock cannot be sent to joblib process pool.
        self._lock = RWLockFair()

    def load_state(
        self,
        survivors: Iterable[tuple[Hashable, S]],
        top_k: Iterable[tuple[Hashable, S]],
    ) -> None:
        self._survivors.clear()
        self._top_k.clear()
        for trace, score in survivors:
            self._survivors.append(Sample(
                self._portable._distr,
                trace,
                self._portable._distr.sample_by_trace(trace),
                score,
            ))
        for trace, score in top_k:
            self._top_k.add(Sample(
                self._portable._distr,
                trace,
                self._portable._distr.sample_by_trace(trace),
                score,
            ))

    @abstractmethod
    def random_new_traces(self, n: int) -> list[Hashable]:
        pass

    def search(self, n_iter: int) -> bool:
        traces = self.random_new_traces(n_iter)
        parallel = Parallel(n_jobs=self._n_jobs, return_as='generator')
        for sample in joblib_imap_unordered(parallel, self._portable.sample_by_trace, traces):
            survived = True
            with self._lock.gen_wlock():
                to_kill = list[tuple[int, Sample[V, S]]]()
                if self._greater_score_fn is not None:
                    for i, sample_ in enumerate(self._survivors):
                        if self._greater_score_fn(sample_.score, sample.score):
                            survived = False
                            break
                        if self._greater_score_fn(sample.score, sample_.score):
                            to_kill.append((i, sample))
                if survived:
                    for i, _ in reversed(to_kill):
                        del self._survivors[i]
                    self._survivors.append(sample)
                in_top_k = self._top_k.add(sample)
            should_stop = False
            if survived and self._survivors_callback is not None:
                should_stop = self._survivors_callback(sample, self.survivors())
                if self._outsider_callback is not None:
                    for _, s in to_kill:
                        self._outsider_callback(s, False)
            if not survived and self._outsider_callback is not None:
                should_stop = self._outsider_callback(sample, True)
            if in_top_k and self._top_k_callback is not None:
                self._top_k_callback(sample, self.top_k())
            if should_stop:
                return False
        return True

    def _sample_weak_cmp(self, s1: Sample[V, S], s2: Sample[V, S]) -> int:
        if self._weak_greater_score_fn is None:
            return 0
        if self._weak_greater_score_fn(s1.score, s2.score):
            return -1
        if self._weak_greater_score_fn(s2.score, s1.score):
            return 1
        return 0

    def survivors(self) -> list[Sample[V, S]]:
        with self._lock.gen_rlock():
            if self._weak_greater_score_fn is not None:
                return sorted(self._survivors, key=functools.cmp_to_key(self._sample_weak_cmp))
            return list(self._survivors)

    def top_k(self) -> list[Sample[V, S]]:
        with self._lock.gen_rlock():
            return self._top_k.data()


class RandomizedSearch(Generic[V, S], BaseRandomizedSearch[V, S]):

    def __init__(
        self,
        distr: BaseDistribution[V],
        score_fn: Callable[[V], S],
        greater_score_fn: Callable[[S, S], bool] | None,
        weak_greater_score_fn: Callable[[S, S], bool] | None = None,
        top_k_kept: int = 0,
        n_jobs: int | None = None,
        sample_callback: Callable[[Sample[V, None]], None] | None = None,
        survivors_callback: \
            Callable[[Sample[V, S], list[Sample[V, S]], set[Hashable]], bool] | None = None,
        outsider_callback: Callable[[Sample[V, S], bool, set[Hashable]], bool] | None = None,
        top_k_callback: Callable[[Sample[V, S], list[Sample[V, S]]], None] | None = None,
    ) -> None:
        def survivors_callback_(sample: Sample[V, S], survivors: list[Sample[V, S]]) -> bool:
            self._trace_history.add(sample.trace)
            if survivors_callback is not None:
                return survivors_callback(sample, survivors, self.trace_history())
            return False
        def outsider_callback_(sample: Sample[V, S], is_new: bool) -> bool:
            if is_new:
                self._trace_history.add(sample.trace)
            if outsider_callback is not None:
                return outsider_callback(sample, is_new, self.trace_history())
            return False
        super().__init__(
            distr,
            score_fn,
            greater_score_fn,
            weak_greater_score_fn,
            top_k_kept,
            n_jobs,
            sample_callback,
            survivors_callback_,
            outsider_callback_,
            top_k_callback,
        )
        self._trace_history = set[Hashable]()

    def load_state(
        self,
        survivors: Iterable[tuple[Hashable, S]],
        top_k: Iterable[tuple[Hashable, S]],
        trace_history: set[Hashable],
    ) -> None:
        BaseRandomizedSearch.load_state(self, survivors, top_k)
        self._trace_history.clear()
        for trace in trace_history:
            self._trace_history.add(trace)

    def random_new_traces(self, n: int) -> list[Hashable]:
        traces = list[Hashable]()
        while True:
            if len(traces) == n:
                break
            if len(self._trace_history) + len(traces) == self._portable._distr.size():
                break
            trace = self._portable._distr.random_trace()
            if trace not in self._trace_history and trace not in traces:
                traces.append(trace)
        return traces

    def trace_history(self) -> set[Hashable]:
        with self._lock.gen_rlock():
            return set(self._trace_history)


class ExhaustiveSearch(Generic[V, S], BaseRandomizedSearch[V, S]):

    def __init__(
        self,
        distr: BaseDistribution[V],
        score_fn: Callable[[V], S],
        greater_score_fn: Callable[[S, S], bool] | None,
        weak_greater_score_fn: Callable[[S, S], bool] | None = None,
        top_k_kept: int = 0,
        n_jobs: int | None = None,
        sample_callback: Callable[[Sample[V, None]], None] | None = None,
        survivors_callback: Callable[[Sample[V, S], list[Sample[V, S]], int], bool] | None = None,
        outsider_callback: Callable[[Sample[V, S], bool, int], bool] | None = None,
        top_k_callback: Callable[[Sample[V, S], list[Sample[V, S]]], None] | None = None,
    ) -> None:
        if (d_s := self._portable._distr.size()) is None:
            raise ValueError
        def survivors_callback_(sample: Sample[V, S], survivors: list[Sample[V, S]]) -> bool:
            self._count += 1
            if survivors_callback is not None:
                return survivors_callback(sample, survivors, self.count())
            return False
        def outsider_callback_(sample: Sample[V, S], is_new: bool) -> bool:
            if is_new:
                self._count += 1
            if outsider_callback is not None:
                return outsider_callback(sample, is_new, self.count())
            return False
        super().__init__(
            distr,
            score_fn,
            greater_score_fn,
            weak_greater_score_fn,
            top_k_kept,
            n_jobs,
            sample_callback,
            survivors_callback_,
            outsider_callback_,
            top_k_callback,
        )
        self._d_s = d_s
        self._count = 0

    def load_state(
        self,
        survivors: Iterable[tuple[Hashable, S]],
        top_k: Iterable[tuple[Hashable, S]],
        count: int,
    ) -> None:
        BaseRandomizedSearch.load_state(self, survivors, top_k)
        self._count = count

    def random_new_traces(self, n: int) -> list[Hashable]:
        if self._count + n > self._d_s:
            raise IndexError
        return [self._portable._distr.trace_by_index(self._count + i) for i in range(n)]

    def count(self) -> int:
        with self._lock.gen_rlock():
            return self._count


@dataclass
class Sample(Generic[V, S]):

    distr: BaseDistribution[V]
    trace: Hashable
    value: V
    score: S

    def explain(self) -> str:
        return self.distr.explain_trace(self.trace)


class SortedList(Generic[T]):

    def __init__(
        self,
        iterable: Iterable[T] | None = None,
        key: Callable[[T], Any] | None = None,
        capacity: int = -1,
    ) -> None:
        self._inner = SortedListBase(iterable, key)
        self._key = key if key is not None else lambda _: 0
        self._capacity = capacity

    def data(self) -> list[T]:
        return list(self._inner)

    def add(self, value: T) -> bool:
        if self._capacity < 0 or len(self._inner) < self._capacity:
            self._inner.add(value)
            return True
        if len(self._inner) == 0 or self._key(value) >= self._key(typing.cast(T, self._inner[-1])):
            return False
        self._inner.add(value)
        self._inner.pop()
        return True

    def clear(self) -> None:
        self._inner.clear()


__all__ = [
    'RandomizedSearch',
    'ExhaustiveSearch',
    'Sample',
]



from __future__ import annotations

from typing import Any

from river.base import BinaryDriftDetector, DriftDetector

from musc.service.concepts.a import BaseDriftDetector, BaseDriftDetectorWithNcl, DriftDetected, Metric


class DriftDetectorByRiver(BaseDriftDetector):

    def __init__(self, drift_detector: DriftDetector, metric: Metric) -> None:
        self._drift_detector = drift_detector
        self._metric = metric

    def step(self, x: Any, y: Any, y_pred: Any) -> DriftDetected | None:
        del x
        self._drift_detector.update(self._metric(y=y, y_pred=y_pred))
        return DriftDetected(None) if self._drift_detector.drift_detected else None

    def clone_without_state(self) -> DriftDetectorByRiver:
        return DriftDetectorByRiver(self._drift_detector.clone(), self._metric)


class DriftDetectorByRiverBinary(BaseDriftDetector):

    def __init__(self, drift_detector: BinaryDriftDetector) -> None:
        self._drift_detector = drift_detector

    def step(self, x: Any, y: Any, y_pred: Any) -> DriftDetected | None:
        del x
        self._drift_detector.update(y != y_pred)
        return DriftDetected(None) if self._drift_detector.drift_detected else None

    def clone_without_state(self) -> DriftDetectorByRiverBinary:
        return DriftDetectorByRiverBinary(self._drift_detector.clone())


class DriftDetectorByRiverWithNcl(BaseDriftDetectorWithNcl):

    def __init__(self, drift_detector: DriftDetector, metric: Metric) -> None:
        assert hasattr(drift_detector, 'new_concept_length')
        self._drift_detector = drift_detector
        self._metric = metric

    def step(self, x: Any, y: Any, y_pred: Any) -> DriftDetected | None:
        del x
        self._drift_detector.update(self._metric(y=y, y_pred=y_pred))
        ncl = self._drift_detector.new_concept_length  # type: ignore
        return DriftDetected(ncl) if ncl is not None else None

    def clone_without_state(self) -> DriftDetectorByRiverWithNcl:
        return DriftDetectorByRiverWithNcl(self._drift_detector.clone(), self._metric)


class DriftDetectorByRiverBinaryWithNcl(BaseDriftDetectorWithNcl):

    def __init__(self, drift_detector: BinaryDriftDetector) -> None:
        assert hasattr(drift_detector, 'new_concept_length')
        self._drift_detector = drift_detector

    def step(self, x: Any, y: Any, y_pred: Any) -> DriftDetected | None:
        del x
        self._drift_detector.update(y != y_pred)
        ncl = self._drift_detector.new_concept_length  # type: ignore
        return DriftDetected(ncl) if ncl is not None else None

    def clone_without_state(self) -> DriftDetectorByRiverBinaryWithNcl:
        return DriftDetectorByRiverBinaryWithNcl(self._drift_detector.clone())


__all__ = [
    'DriftDetectorByRiver',
    'DriftDetectorByRiverBinary',
    'DriftDetectorByRiverWithNcl',
    'DriftDetectorByRiverBinaryWithNcl',
]



from __future__ import annotations

from abc import ABC, abstractmethod
from threading import Event, Lock


class BaseActiveCount(ABC):

    @abstractmethod
    def inc(self) -> None:
        pass

    @abstractmethod
    def dec(self) -> None:
        pass


class ActiveCount(BaseActiveCount):

    def __init__(self, count: int, idle_event: Event, lock: Lock) -> None:
        self._count = count
        self._idle_event = idle_event
        self._lock = lock

    @staticmethod
    def create() -> tuple[ActiveCount, Event]:
        idle_event = Event()
        idle_event.set()
        return ActiveCount(0, idle_event, Lock()), idle_event

    def inc(self) -> None:
        with self._lock:
            self._count += 1
            if self._count == 1:
                self._idle_event.clear()

    def dec(self) -> None:
        with self._lock:
            if self._count == 0:
                raise RuntimeError
            self._count -= 1
            if self._count == 0:
                self._idle_event.set()


class NullActiveCount(BaseActiveCount):

    def inc(self) -> None:
        pass

    def dec(self) -> None:
        pass


__all__ = [
    'BaseActiveCount',
    'ActiveCount',
    'NullActiveCount',
]
